/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *
 *  Revision History:
 *  06/16/00  release 0.3 (split from the CCanvas). 
 *
 *
 */

package	adm;

import	java.io.*;   
import	java.awt.*;

import  cdev.*;
import  cdev.data.*;


/**
 * CAnimatedCanvas class extends CCanvas and implements double buffering. 
 * It can be used as a base class for widgets with complicated graphics. 
 * The background paint and foreground paint are separated in this class. 
 * The background painted is only called the first time in the paint method. 
 * Only the foreground paint is performed for each update. This provides 
 * a better image update (smoother) and an efficient paint. It is better to 
 * use this base class when developing widgets with the complicated
 * backgrounds or dynamic graphics. 
 * <BR><BR>
 * CAnimatedCanvas is abstract class. The derived classes must implement 
 * paintBackground and paintForeground methods.
 *
 *
 * @version   0.2  04/15/00
 * @author    Ying Chen, Chip Watson
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.3
 */

public abstract class CAnimatedCanvas extends CCanvas {

    static final int MINSIZE = 5;

    //--- define default value --------------------------
    protected int colorMode  = CLRMOD_STATIC;
    protected int labelStyle = LABEL_NONE;
    protected int rangeStyle = RANGE_FROM_PROPERTY;
    protected int frameW = 2;           // dafault face frame = 2 pixels
    protected double controlHi = 100.0;
    protected double controlLo = 0.0;
    protected double displayHi = 90.0;
    protected double displayLo = 10.0;
    protected double alarmHi = 80.0;
    protected double alarmLo = 20.0;
    protected double warningHi = 70.0;
    protected double warningLo = 30.0;
    protected double rangeHi = displayHi;
    protected double rangeLo = displayLo;
    protected double precision = 1.0;
    protected Color fgColor = Color.green;
    protected Color bgColor = Color.gray;
    protected Font font =  new Font("Dialog", Font.PLAIN, 12);
    
    //--- boolean flag -------------------------------------------
    protected   boolean firstTime = true;   
    protected   boolean firstCall = true; 
    protected   boolean gotLimits = false;

    //--- used for double-buffering ------------------------------
    protected Dimension offDimension;     // double-buffer size
    protected Image     offImage;         // off-screen buffer
    protected Graphics  offGraphics;      // off-screen paint handle
  
    protected double currVal = 0;
    protected Color currColor = null;


    /**
     * Null Constructor
     */
    public CAnimatedCanvas() 
    {
   	this(0, 0);
    }

    /**
     * Constructor with device name and property name
     * 
     * @param dName    device name
     * @param pName    property name
     */
    public CAnimatedCanvas(String dName, String pName) 
    {
	setSize (0, 0);
	setDeviceAndPropertyName(dName, pName);
    }
    
    /**
     * Constructor
     * Defines the canvas size and uses default foreground and 
     * background colors.
     * @param width    component width
     * @param height   component height
     */
    public CAnimatedCanvas(int width, int height) 
    {
	setSize (width, height);
    }
    

    /**
     * Constructor
     * Defines the canvas size and sets foreground and background colors.
     * @param width     component width
     * @param height    component height
     * @param fg        foreground color
     * @param bg        background color
     */
    public CAnimatedCanvas( int width, int height, Color fg, Color bg )
    {
	this(width, height);
        if (fg != null) fgColor = fg;
        if (bg != null) bgColor = bg;
    }


    /**
     * Set canvas size. 
     * @param newWidth     width of the component
     * @param newHeight    height of the component
     */
    public void setSize (int newWidth, int newHeight)
    {
	int w = Math.max(newWidth, MINSIZE);
	int h = Math.max(newHeight, MINSIZE);
	super.setSize(w, h);
	offGraphics = null;      // clear the offGraphics to force a resize later
    }


    /**
     * Set canvas size. 
     * @param d    dimension of the component
     */
    public void setSize(Dimension d)
    {
	this.setSize(d.width, d.height);
    }
   

    /** 
     * Set the bounds of the canvas.
     * @param x    component X location
     * @param y    component Y location
     * @param w    component width
     * @param h    component height
     */
    public void setBounds(int x, int y, int w, int h)
    {
	super.setBounds(x,y,w,h);
	offGraphics = null;	// clear the offGraphics to force a resize later
    }


   /** 
     * Set the bounds of the canvas.
     * @param r    a rectange bounding box for component
     */
    public void setBounds(Rectangle r)
    {
	super.setBounds(r);
	offGraphics = null;	// clear the offGraphics to force a resize later
    } 


    /**
     * Set text font for labeling and displaying values.
     * @param newFont    font for any text on the component
     */
    public void setFont (Font newFont)  {
	font = newFont;
	offGraphics = null;	// force a repaint of the background
    }


    /**
     * Return the font of the text on the component
     */
    public Font getFont () { return font;}
  

    /**
     * Set component background color
     * @param newBg    the background color of component
     */
    public void setBackground (Color newBg) {
	bgColor = newBg;
	offGraphics = null;	// force a repaint of the background
    }
 

   /**
     * Set compoenet foreground color
     * @param newFg   the foreground color of component
     */
    public void setForeground (Color newFg) { 
	fgColor = newFg;
	offGraphics = null;	// force a repaint of the background
    }


    /**
     * Return the background color of the component
     */
    public Color getBackground () { return bgColor; }


    /**
     * Return the foreground color of the component
     */
    public Color getForeground () { return fgColor; }

 
    /**
     * Set the frame width of component face
     * @param pixels   the pixels of the frame width
     */
    public void setFrameWidth (int pixels) 
    {
       frameW = pixels;
       offGraphics = null;	// force a repaint of the background and foreground
    }


    /**
     * Return the frame width of component face
     */
    public int getFrameWidth () 	   { return frameW; }


    /**
     * Set the label style property
     * @param newLab   the new label style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.LABEL_NONE
     *             <LI>  CConstants.LABEL_AXIS
     *             <LI>  CConstants.LABEL_VALUE
     *             <LI>  CConstants.LABEL_ALL
     *             </OL>
     */
    public void setLabelStyle (int newLab)
    {
	if (labelStyle != newLab) {
	    labelStyle = newLab;	    
	    offGraphics = null;
	}
    }
    

    /**
     * Return the label style property
     */
    public int getLabelStyle () { return labelStyle; }
  
  
    /**
     * Set the color mode property
     * @param newColorMode   the new color mode, which may be one of:
     *             <OL>
     *             <LI>  CConstants.CLRMOD_STATIC
     *             <LI>  CConstants.CLRMOD_ALARM
     *             </OL>
     */
    public void setColorMode (int newColorMode) {
	if (colorMode != newColorMode) {
	    colorMode = newColorMode;
	    offGraphics = null;
	}
    }
    
    
    /**
     * Return the color mode property
     */
    public int getColorMode () { return colorMode; }
    
    
    /**
     * Set the range style property
     * @param newRangeStyle  the new range style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.RANGE_AUTO
     *             <LI>  CConstants.RANGE_FROM_PROPERTY
     *             <LI>  CConstants.RANGE_STATIC
     *             </OL>
     */
    public void setRangeStyle (int newRangeStyle) {
	if (rangeStyle != newRangeStyle) {
	    rangeStyle = newRangeStyle;
	    offGraphics = null;
	}
    }


    /**
     *  Return the range style property
     */
    public int getRangeStyle () { return rangeStyle; }

 
    /**
     * Set the display limits. The range style property
     * will be forced to RANGE_STATIC.
     * 
     * @param lo   the display low
     * @param hi   the display high
     *
     * hi must be larger than lo otherwise nothing is changed.
     */
    public void setDisplayRange (double lo, double hi)
    {
	if (hi - lo > 0) {
	    displayHi = hi;
	    displayLo = lo;
	    rangeStyle = RANGE_STATIC;
	    rangeHi = displayHi;
	    rangeLo = displayLo;
	    offGraphics = null;
	}
    }

    /**
     * Set the display high. The range style property
     * will be forced to RANGE_STATIC.
     * 
     * @param hi   the display high
     *
     * hi must be larger than displayLo otherwise nothing is changed.
     */
    public void setDisplayHi (double hi)
    {
	if (hi - displayLo > 0) {
	    displayHi = hi;
	    rangeHi = displayHi;
	    rangeStyle = RANGE_STATIC;
	    offGraphics = null;
	}
    }

    /**
     * Set the display low. The range style property
     * will be forced to RANGE_STATIC.
     * 
     * @param lo   the display low
     *
     * lo must be smaller than displayHi otherwise nothing is changed.
     */
    public void setDisplayLo (double lo)
    {
	if (lo - displayHi < 0) {
	    displayLo = lo;
	    rangeLo = displayLo;
	    rangeStyle = RANGE_STATIC;
	    offGraphics = null;
	}
    }

    /**
     * Return display high
     */
    public double getDisplayHi () { return displayHi; }


    /**
     * Return display low
     */
    public double getDisplayLo () { return displayLo; }


    //--------------------------------------------------------------
    // child class needs to overwrite these two methods 
    //--------------------------------------------------------------
    protected void paintBackground (Graphics g) {}
    protected void paintForeground (Graphics g) {}


    /**
     * Set up offscreen graphics.  This is needed for both 
     * a new screen paint and a resize paint, or anything which
     * changes the widget's background (color, font, etc.).
     */
    protected void setupOffscreen() {
        offDimension = getSize();
	//System.out.println ("set new offDim size["+offDimension+"]...");
        offImage = createImage(offDimension.width, offDimension.height);
        offGraphics = offImage.getGraphics();
        offGraphics.setColor( bgColor );
        offGraphics.fillRect(0, 0, offDimension.width, offDimension.height);
    }


    /**
     * Overwrite the parent update method so it only call this.paint(g)
     */
    public final void update(Graphics g)
    {
	paint(g);
    }  


    /**
     * Draw offscreen graphics and repaint it onto the screen.
     * Calls derived class' paintForeground(g) and paintBackground(g).
     */
    public final void paint(Graphics g)
    {
	if (firstTime) {
    	    firstTime = false;
	    //System.out.println ("CCanvas::paint: first time paint.");
	}
        if (isShowing() && gotLimits) {
	    if ( offGraphics == null ) {
		setupOffscreen();
		paintBackground (offGraphics);
	    }
	    paintForeground (offGraphics);
	    g.drawImage(offImage, 0, 0, null);
	}
    }  


   public void deviceChanged(DeviceEvent evt) 
   { 
       boolean assumeGotLimits = true;
       if (firstCall) {
	  firstCall = false;

	  //--- get teh display range --------------------------
	  try {
	      currVal=evt.getData().getDataEntry("value").doubleValue(); 
	  }
	  catch (Exception e) { System.out.println("Failed to get first value"); }
	  switch (rangeStyle) {
	  case RANGE_STATIC:
	      break;
	  case RANGE_AUTO:
	      displayHi = currVal + 0.1*Math.abs(currVal);
	      displayLo = currVal - 0.1*Math.abs(currVal);
	      controlHi = currVal + 0.1*Math.abs(currVal);
	      controlLo = currVal - 0.1*Math.abs(currVal);
	      break;
	  default:    //case RANGE_FROM_PROPERTY:
	      try {
                  displayHi=evt.getData().getDataEntry("displayHigh").doubleValue();
                  displayLo=evt.getData().getDataEntry("displayLow").doubleValue();
	      }
              catch (Exception e) {
		  firstCall = true;
		  assumeGotLimits = false;
		  try { device.get(propertyName, this); }
		  catch (Exception ee) { ee.printStackTrace(); }
		  System.out.println("Failed to get display range");
	      }
	      break;
	  }
	  rangeHi = displayHi;
	  rangeLo = displayLo;

	  //--- get the alarm and warning range ----------------------
	  try {
	      alarmHi=evt.getData().getDataEntry("alarmHigh").doubleValue();
	      alarmLo=evt.getData().getDataEntry("alarmLow").doubleValue();
	  }
	  catch  (Exception e) {
	      alarmHi = displayHi;
	      alarmLo = displayLo;
	      System.out.println("Failed to get alarm range");
	  }
	  try {
	      warningHi=evt.getData().getDataEntry("warningHigh").doubleValue();
	      warningLo=evt.getData().getDataEntry("warningLow").doubleValue();
	  }
	  catch  (Exception e) {
	      warningHi = displayHi;
	      warningLo = displayLo;
	      System.out.println("Failed to get warining range");
	  }

	  //--- get the control range and precision -----------------------
	  try {
	      controlHi=evt.getData().getDataEntry("controlHigh").doubleValue();
	      controlLo=evt.getData().getDataEntry("controlLow").doubleValue();
	  }
	  catch (Exception e) { 
	      controlHi = displayHi;
	      controlLo = displayLo;
	      System.out.println("Failed to get control range");
	  }
	  //System.out.println("CCanvas::deviceChanged alarmHi="+alarmHi); 
	  //System.out.println("CCanvas::deviceChanged warningHi="+warningHi);
	  //System.out.println("CCanvas::deviceChanged precision="+precision);
	  try {
	      precision=evt.getData().getDataEntry("precision").doubleValue();
	  }
	  catch ( Exception e) {
	      precision = (controlHi - controlLo)/100;
	      System.out.println("Use default precision ("+precision+")");
	  }
	  gotLimits = assumeGotLimits;
	  if (gotLimits)  startMonitor();

      }
      //--- this will be call for every update ------------         
      try { currVal = evt.getData().getDataEntry("value").doubleValue();} 
      catch (Exception e) { e.printStackTrace(); }

      //--- find out the currColor used in paint ---------
      if (colorMode == CLRMOD_STATIC) { currColor = fgColor; }
      else {
         if ((currVal-alarmHi > 0) || (currVal-alarmLo < 0)) 
	    currColor = alarmColor;
	 else if ((currVal-warningHi > 0) || (currVal-warningLo < 0)) 
	    currColor = warningColor;
         else 
	    currColor = fgColor;
      }
      repaint();
   }		

}  // end CAnimatedCanvas






