/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  04/17/00:  Release version 0.2.
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import cdev.*;
import cdev.data.*;


/**
 * CChoice is subclass of awt.Choice. It is used to control a device/property 
 * value of enumerated type. If the connected device/property is not an 
 * Enum data type, three items "Invalid", deviceName and propertyName will 
 * be shown in the choice. The selection is fixed on the first item (Invalid). 
 * <BR><BR>
 * The state change event action is built into this widget. When the item 
 * state is changed, the cursor will change to the waiting cursor and a set 
 * action will be sent to the control system. If the set fails, the selected 
 * item (display) will revert to the old one and an error message will be 
 * printed to standard output. The selection is also disabled.
 *
 *
 * @version   0.2  03/29/00
 * @author    Ying Chen, Chip Watson 
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.2
 *
 */
public class CChoice extends Choice 
    implements CInterface, DeviceListener, ItemListener {

    static final String BLANK = "";  
    static final int VAL_TAG = Data.convertKey("value");  
    private static DeviceContext defContext = new DeviceContext();
    private static DeviceContext limitsContext;
    static {
	Data limitsContextData = new Data();
	limitsContextData.insert("displayHigh", 1);
	limitsContext = new DeviceContext(limitsContextData);
    }
    private Data tdata = new Data();

    private Device device = null;
    private String deviceName = null;
    private String propertyName = null;
    private String lastDeviceName = null;
    private String lastPropertyName = null;

    //--- boolean flag -------------------------------------------
    private boolean active = false;
    private boolean firstCall = true; 
    private boolean gotTags = false;

    private String value; 
    private String[] tags;
    private DeviceListener setListener = null;
    private Cursor defCur  = new Cursor(Cursor.DEFAULT_CURSOR);
    private Cursor waitCur = new Cursor(Cursor.WAIT_CURSOR);


    /**
     * Null constructor. 
     */
    public CChoice()  {
	addItemListener(this);
    }
    
    /**
     * Null constructor  with device name and property name
     *
     * @param dName    device name
     * @param pName    property name
     */
    public CChoice(String dName, String pName)  {
	this();
	setDeviceName(dName);
	setPropertyName(pName);
    }


     /**
     * Set limitsContext to get all meta-data at first time
     */
    synchronized protected void getFirstValue () 
    {
       try {
	  if (deviceName == null || propertyName == null)   return;
          if (!deviceName.equals(lastDeviceName)) {
             device = new Device(deviceName);
             lastDeviceName = deviceName;
             }
          device.setContext (limitsContext);
	  device.get(propertyName, this);
          firstCall = true; 
          }
       catch(Exception e) { e.printStackTrace(); }
     }
     

    /**
     * Set the Device name 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setDeviceName(String newName) 
    {
       newName = CUtility.makeValidName(newName);
       if (!CUtility.testNameMatch(deviceName, newName)) {
          stopMonitor();
          deviceName = newName;
          getFirstValue ();
       } 
    }
    

    /**
     * Return the device name to which this component is connected.
     * If device name is not set, return empty string "". 
     */
    public String getDeviceName()
    {
       return (deviceName==null)?BLANK:deviceName;
    }


    /**
     * Set device property name 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setPropertyName(String newName)
    {
	newName = CUtility.makeValidName(newName);
        if (!CUtility.testNameMatch(propertyName, newName)) {
           stopMonitor();
           propertyName = newName; 
           getFirstValue ();        
        }
    }


    /**
     * Return the device property name to which this component is connected.
     * If property name is not set, return empty string "". 
     */
    public String getPropertyName()
    {
	return (propertyName==null)?BLANK:propertyName;
    }
	

   /**
     * Set device name and device property name
     * @param newDeviceName    the device name you want to connect
     * @param newPropertyName  the device property name you want to connect
     */  
    synchronized public void setDeviceAndPropertyName
                                       (String newDeviceName,
					String newPropertyName)
    {
	newDeviceName = CUtility.makeValidName(newDeviceName);
	newPropertyName = CUtility.makeValidName(newPropertyName);
        if ((!CUtility.testNameMatch(deviceName, newDeviceName)) ||
           (!CUtility.testNameMatch(propertyName, newPropertyName))) {
           stopMonitor();
	   deviceName = newDeviceName;
           propertyName = newPropertyName; 
           getFirstValue ();        
        }
    }


    protected void stopMonitor() 
    {
        if (active) {
          try {
	     active = false;
	     device.monitorOff(propertyName,this);
	  } 
	  catch (IOException e) {
	     e.printStackTrace();
	  }
       }
    }
	
    protected void startMonitor() 
    {
	try {
	    if (active) stopMonitor();
	    if (deviceName == null || propertyName == null) return;
	    if (!deviceName.equals(lastDeviceName)) {
		device = new Device(deviceName);
		lastDeviceName = deviceName;
	    }    
	    device.setContext (defContext);
	    device.monitorOn(propertyName,this);
	    lastPropertyName = propertyName;
	    active = true;
	} 
	catch (IOException e) {
	    active = false;
	    e.printStackTrace();
	}
    }
    
    /**
     * Get the Enum value from control system and add to the choice
     * at first call. Change the select item to the current Enum
     * when the control system updates.
     */ 
    public void deviceChanged(DeviceEvent event) {
	if (firstCall) {   // from get call
	    this.removeAll();
	    try {
		DataEntry de = event.getData().getDataEntry("value");
		if (de.getValue() instanceof String[]) {
		    tags = de.stringArray();
		    for (int i=0; i< tags.length; i++) {
			this.add(tags[i]);
			System.out.println("array["+i+"] = "+tags[i]);
		    }	
		    firstCall = false;
		    startMonitor();
		    gotTags = true;
		}
		else { 
		    this.add("Invalid");
		    this.add(deviceName);
		    this.add(propertyName);
		    System.out.println("Invalid "+deviceName+"."+propertyName);
		    System.out.println("data type is not ENUM");
		    firstCall = false;
		}
	    }
	    catch (Exception e) { 
		System.out.println("failed to Enun string array");
		e.printStackTrace(); 
	    }   
	}
	else {      // from monitorOn 
	    try {
		value = event.getData().getDataEntry("value").stringValue();
		System.out.println("CChoice:: new value="+value);
		System.out.println("new value = "+value);
		this.select(value);
	    } catch (Exception e) {
		System.out.println("failed to get new value");
		e.printStackTrace();
	    } 
	}
    }
    
    /**
     * To send the new value to the control system. The cursor change to
     * waiting cursor. The setListener will keep track the error message
     * return from the control system and set the cursor back to default
     * cursor.
     */
    public void itemStateChanged(ItemEvent evt) {
	if (setListener==null) setListener = new DeviceListener() {
	    public void deviceChanged(DeviceEvent evt) {
		if (evt.getError() == null) {
		    value = getSelectedItem();	    
		    System.out.println("success to set value ("+value+")");
		    setCursor(defCur);
		} else {
		    select(value);
		    System.out.println("failed to set the value");
		    setCursor(defCur);
		}
	    }
	};
	
	if (gotTags) {
	    int newValue = getSelectedIndex();
	    tdata.insert(VAL_TAG, tags[newValue]);
	    try {
		//System.out.println(tdata);
		setCursor(waitCur);
		device.set (propertyName, tdata, setListener);
		device.getControlSystem().flush();
	    } catch (Exception ex) {
		System.out.println ("failed to send value ("+newValue+")");
	    }
	}
	else {
	    select(0);
	}
    }
    
}  // end of CChoice










