/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *
 * Revision History:
 *   06/16/00:  release 0.3 (initial version).
 *   08/29/00:  release 0.4 (add class query and property query).
 *
 */

package adm;

import java.util.Vector;
import adm.CConstants;
import adm.event.*;
import cdev.*;
import cdev.data.*;
import cdev.dir.*;

/**
 * CDirectory is an invisible bean used to query the control system for 
 * lists of devices, properties, or classes. It is an introspection 
 * widget -- it doesn't monitor any value in the control system. 
 * <BR><BR>
 * This bean is configured by specifing a regular expression for one or 
 * more of class name, device name, or property name. Depending upon a 
 * flag set by the user, an array of device names, property names or class 
 * names can be retrieved. The property name retrived are all the properties 
 * for the class. The property expression pattern match 
 * (set/getPropertyExpression method) will not be included in this release 
 * (0.4). 
 * <BR><BR>
 * The setQueryType method must be called after set*Expression to trigger 
 * the DirectoryEvent. The query type may be one of: CLASS_QUERY, DEVICE_QUERY 
 * or PROPERTY_QUERY defined in CConstants class. When the query type is set 
 * to CLASS_QUERY or PROPERTY_QUERY, the classExpression must be set using 
 * setClassExpression. When the query type is DEVICE_QUERY, either 
 * classExpression or deviceExpression need be set. 
 * <BR><BR>
 * To receive the directory service response, an object must implement the
 * CDirectoryListener interface and add itself as a listener to this bean. 
 * There is one method, receivedNames (CDirectoryEvent e), in the 
 * CDirectoryListener interface. This method takes CDirectoryEvent as an 
 * argument. By calling getNames() in the CDirectoryEvent class, the name 
 * array (device names, property names or class names) can be obtained. 
 *
 *
 * @version   0.4  09/17/00
 * @author    Ying Chen, Chip Watson
 *            
 * @since     adm0.3
 *
 */
public class CDirectory implements Runnable {
    
    private String classQuery = null;
    private String deviceQuery = null;
    private String propertyQuery = null;
    private int queryType = -1;
    private String[] nameString = null;

    private Thread worker = null;
    private Vector allListeners = new Vector();

    /**
     * Null Constructor
     */
    public CDirectory () {  }

    /**
     * Set the class expression 
     * @param s     the regular expression for class name
     */
    public void setClassExpression (String s) { classQuery = s; }

    /**
     * Return class expression. If not set, return null. 
     */
    public String getClassExpression ()  { return classQuery; }

    /**
     * Set the device expression 
     * @param s     the regular expression for device name
     */
    public void setDeviceExpression (String s) { deviceQuery = s; }

    /**
     * Return device expression. If not set, return null. 
     */
    public String getDeviceExpression ()     { return deviceQuery; }

    /**
     * Set the property expression 
     * @param s     the regular expression for property name
     */
    public void setPropertyExpression (String s) { propertyQuery = s; }

    /**
     * Return property expression. If not set, return null. 
     */
    public String getPropertyExpression ()   { return propertyQuery; }


    /**
     * Set the queryType. It must be called after the expressions are set.
     * @param type_id   the query type, which may be one of:
     *                  <OL>
     *                  <LI>  CConstants.DEVICE_QUERY
     *                  <LI>  CConstants.PROPERTY_QUERY 
     *                  <LI>  CConstants.CLASS_QUERY
     *                  </OL>
     */
    public void setQueryType (int type_id) { 
	queryType = type_id; 
	start();
    }
    
    /**
     * Return query type. If not set, return -1. 
     */
    public int getQueryType ()  { return queryType; }

    /**
     * Start a new thread
     */
    private void start () {
	if (queryType == -1) {
	    System.out.println("queryType must be set"); 
	    return;
	}
        if (worker == null || !worker.isAlive()) {
            worker = new Thread(this);
            worker.start();
        }
    }

    /**
     * Process the directory service and notify the listener
     */
    public void run () {  
	try {      
	    DirectoryService ds = Config.getDirectoryService();
	    if ((queryType == CConstants.CLASS_QUERY) && 
		(classQuery != null)) {
		try {
		    nameString = ds.getDeviceNames("class="+classQuery);
		    notifyDirectoryListener ();
		} catch(Exception e) { 
		    System.out.println("(no device match class '"+classQuery+"') "+e);
		}
	    }
	    else if (queryType == CConstants.DEVICE_QUERY) {
		try {
		    if (deviceQuery != null) {
			nameString = ds.getDeviceNames("device="+deviceQuery);
			notifyDirectoryListener ();
		    }
		    if (classQuery != null) {
			nameString = ds.getDeviceNames("class="+classQuery);
			notifyDirectoryListener ();
		    }
		} catch(Exception e) { 
		    System.out.println("(no devices found "+e);
		}
	    }
	    else if (queryType == CConstants.PROPERTY_QUERY) {
		if (classQuery != null) {
		    try {
			DeviceClass dc = ds.getDeviceClass(classQuery);
			System.out.println("device class =" +dc);			
			nameString = dc.getPropertyNames();
			notifyDirectoryListener ();
		    } catch(Exception e) { 
			System.out.println("failed to get property for class '"
					   +classQuery+"' "+e);
		    }
		}
		else {
		    System.out.println("You must set the class query");
		}
	    }   
	} catch (Exception e) {
	    System.out.println("directory exception: "+e);
	}    
    }

    /**
     * Add the directory listener to receive the notification
     */
    public synchronized void 
	addDirectoryListener (CDirectoryListener dl) {
	if (!allListeners.contains(dl))
	    allListeners.addElement(dl);
    }

    /**
     * Remove the directory listener service
     */
    public synchronized void 
	removeDirectoryListener (CDirectoryListener dl) { 
	if (allListeners.contains(dl))
	    allListeners.removeElement(dl);
    }

    /**
     * Fire the event change - notify all the directory listener
     */
    private void notifyDirectoryListener () {
	CDirectoryEvent de = new CDirectoryEvent(this, queryType, nameString);
	//--- make a copy of the listener vector ---------------
	//--- in case it chnages while we firing events --------
	Vector v;
	synchronized (this) { v = (Vector)allListeners.clone(); }

	//--- fire the event to all listeners -----------------
	for (int i=0; i< v.size(); i++ ) 
	    ((CDirectoryListener) v.elementAt(i)).receivedNames(de);
    }

}



