/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 * Revision History:
 *   04/17/00:  Release version 0.2.
 *   06/24/00: Release version 0.3 (fix some bugs).
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import cdev.*;
import cdev.data.*;


/**
 * CList is subclass of awt.List. It is used to control a device/property 
 * value of enumerated type (Fig.1). If the connected device/property is 
 * not an Enum data type, three items "Invalid", deviceName and
 * propertyName will be shown in the list. The multiple selection of this
 * widget is turn off.
 * <BR><BR>
 * The state change event action is built into this widget. When the item 
 * state is changed, the cursor will change to the waiting cursor and a set
 * action will be sent to the control system. If the set fails, the selected 
 * item (display) will revert to the old one and an error message will be 
 * printed to standard output. The selection is also disabled.
 *
 *
 * @version   0.2  03/29/00
 * @author    Ying Chen, Chip Watson 
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.2
 *
 */
public class CList extends List
    implements CInterface, DeviceListener, ItemListener {
    
    private static int VAL_TAG = Data.convertKey("value");
    private static String BLANK = "";
    private static DeviceContext defContext = new DeviceContext();
    private static DeviceContext limitsContext;
    static {
	Data limitsContextData = new Data();
	limitsContextData.insert("displayHigh", 1);
	limitsContext = new DeviceContext(limitsContextData);
    }
    private Data tdata = new Data();
    private boolean active = false;
    private Device device = null;
    private String deviceName = null;
    private String propertyName = null;
    private String lastDeviceName = null;
    private String lastPropertyName = null;

    private boolean firstCall = true; 
    private boolean gotTags = false;
    private int value;
    private String[] tags;
    private DeviceListener setListener = null;
    private Cursor defCur  = new Cursor(Cursor.DEFAULT_CURSOR);
    private Cursor waitCur = new Cursor(Cursor.WAIT_CURSOR);
    
    /**
     * Null constructor. 
     */
    public CList()  {
	super(2, false);
	addItemListener(this);
    }


    /**
     * Null constructor  with device name and property name
     *
     * @param dName    device name
     * @param pName    property name
     */
    public CList(String dName, String pName)  {
	this();
	setDeviceName(dName);
	setPropertyName(pName);
    }


    
    /**
     * Overwrite the setMultipleMode method to turn off the
     * multiple selection.
     */
    public void setMultipleMode (boolean b) {
	super.setMultipleMode(false);
    }

     /**
     * Set limitsContext to get all meta-data at first time
     */
    synchronized protected void getFirstValue () 
    {
       try {
	  if (deviceName == null || propertyName == null)   return;
          if (!deviceName.equals(lastDeviceName)) {
             device = new Device(deviceName);
             lastDeviceName = deviceName;
             }
          device.setContext (limitsContext);
	  device.get(propertyName, this);
          firstCall = true; 
          }
       catch(Exception e) { e.printStackTrace(); }
     }
     

    /**
     * Set the Device name 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setDeviceName(String newName) 
    {
       newName = CUtility.makeValidName(newName);
       if (!CUtility.testNameMatch(deviceName, newName)) {
          stopMonitor();
          deviceName = newName;
          getFirstValue ();
       } 
    }
    

    /**
     * Return the device name to which this component is connected.
     * If device name is not set, return empty string "". 
     */
    public String getDeviceName()
    {
       return (deviceName==null)?BLANK:deviceName;
    }


    /**
     * Set device property name 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setPropertyName(String newName)
    {
	newName = CUtility.makeValidName(newName);
        if (!CUtility.testNameMatch(propertyName, newName)) {
           stopMonitor();
           propertyName = newName; 
           getFirstValue ();        
        }
    }


    /**
     * Return the device property name to which this component is connected.
     * If property name is not set, return empty string "". 
     */
    public String getPropertyName()
    {
	return (propertyName==null)?BLANK:propertyName;
    }
	

    protected void stopMonitor() 
    {
        if (active) {
          try {
	     active = false;
	     device.monitorOff(propertyName,this);
	  } 
	  catch (IOException e) {
	     e.printStackTrace();
	  }
       }
    }
	
    protected void startMonitor() 
    {
	try {
	    if (active) stopMonitor();
	    if (deviceName == null || propertyName == null) return;
	    if (!deviceName.equals(lastDeviceName)) {
		device = new Device(deviceName);
		lastDeviceName = deviceName;
	    }    
	    device.setContext (defContext);
	    device.monitorOn(propertyName,this);
	    lastPropertyName = propertyName;
	    active = true;
	} 
	catch (IOException e) {
	    active = false;
	    e.printStackTrace();
	}
    }

    /**
     * Get the Enum value from control system and add to the choice
     * at first call. Change the select item to the current Enum
     * when the control system updates.
     */
    public void deviceChanged(DeviceEvent event) {
	if (firstCall) {    // from get call
	    this.removeAll();
	    try {
		DataEntry de = event.getData().getDataEntry("value");
		if (de.getValue() instanceof String[]) {
		    tags = de.stringArray();
		    for (int i=0; i< tags.length; i++) {
			this.add(tags[i]);
			System.out.println("array["+i+"] = "+tags[i]);
		    }	
		    firstCall = false;
		    startMonitor();
		    gotTags = true;
		}
		else { 
		    this.add("Invalid");
		    this.add(deviceName);
		    this.add(propertyName);
		    System.out.println("Invalid "+deviceName+"."+propertyName);
		    System.out.println("data type is not ENUM");
		    firstCall = false;
		}
	    } catch (Exception e) { e.printStackTrace(); }   
	}
	else {    // from monitorOn
	    try {
		String s = event.getData().getDataEntry("value").stringValue();
		value = getTagIndex(s);
		select(value);
	    }
	    catch (Exception e) { e.printStackTrace(); }
	}  
    }
    
    /**
     * To send the new value to the control system. The cursor change to
     * waiting cursor. The setListener will keep track the error message
     * return from the control system and set the cursor back to default
     * cursor.
     */
    public void itemStateChanged(ItemEvent evt) {
	if (setListener==null) setListener = new DeviceListener() {
	    public void deviceChanged(DeviceEvent evt) {
		if (evt.getError() == null) {
		    value = getSelectedIndex();	    
		    System.out.println("success to set value ("+value+")");
		    setCursor(defCur);
		} else {
		    select(value);
		    System.out.println("failed to set the value");
		    setCursor(defCur);
		}
	    }
	};
	if (gotTags) {
	    int newValue = getSelectedIndex();
	    tdata.insert(VAL_TAG, tags[newValue]);
	    try {
		//System.out.println(tdata);
		setCursor(waitCur);
		device.set (propertyName, tdata, setListener);
		device.getControlSystem().flush();
	    } catch (Exception ex) {
		System.out.println ("failed to send value ("+newValue+")");
	    }
	}
	else { select(0); }
    }
       
    private int getTagIndex(String tagName) {
	int index = 0;
        if (gotTags) {
            boolean found = false;
            int i;
            for (i=0; i<tags.length && !found; i++) {
                if (tags[i].equals(tagName)) 
                    found = true;
            }
            index = i-1;
        }
        return index;
    }

}  // end of CList










