/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *
 *  Revision History:
 *  03/10/00:  Release version 0.1.
 *  04/17/00:  Release version 0.2. (no change)
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;

/**
 * CPanel is a subclass of awt.Panel. 
 * It supports recursive setDeviceName, setPropertyName to all contained
 * components which implement CInterface. This makes it possible to build 
 * a small GUI to control (for example) a single device, and then control 
 * the device name for the whole panel from a choice box or list. 
 *
 *
 * @version   0.2  04/17/00
 * @author    Chip Watson, Ying Chen
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.1
 */
public class CPanel extends Panel implements CInterface
{

    private String deviceName = null;
    private String propertyName = null;
    private static String BLANK = "";
    
    /**
     * Null Constructor
     */
    public CPanel()
    {
	//{{INIT_CONTROLS
	setLayout(null);
	setSize(400,300);
	//}}
    }
    
    /**
     * Set the Device name for all components in the panel 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setDeviceName(String newName)
    {
	newName = CUtility.makeValidName(newName);
	boolean same = CUtility.testNameMatch(deviceName,newName);
	
	// if device name has changed, tell all components
	if (!same) {
	    deviceName = newName;
	    Component [] components = getComponents();
	    for (int i=0;i<components.length;i++)
		if (components[i] instanceof CInterface) {
		    CInterface cc = (CInterface) components[i];
		    cc.setDeviceName(deviceName);
		}
	}
    }
    

   /**
     * Return the device name.
     * If device name is not set, return empty string "". 
     */
    public String getDeviceName()
    {
	return (deviceName==null)?BLANK:deviceName;
    }
    

    /**
     * Set device property name for all components in the panel. 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setPropertyName(String newName)
    {
	newName = CUtility.makeValidName(newName);
	boolean same = CUtility.testNameMatch(propertyName, newName);
	
	// if property name has changed, tell all components
	if (!same) {
	    propertyName = newName;
	    Component [] components = getComponents();
	    for (int i=0;i<components.length;i++)
		if (components[i] instanceof CInterface) {
		    CInterface cc = (CInterface) components[i];
		    cc.setPropertyName(propertyName);
		}
	}
    }


     /**
     * Return the device property name.
     * If property name is not set, return empty string "". 
     */
    public String getPropertyName()
    {
	return (propertyName==null)?BLANK:propertyName;
    }
    
    /*
    static public void main(String args[])
    {
	class DriverFrame extends Frame {
	    public DriverFrame() {
		addWindowListener(new WindowAdapter() {
		    public void windowClosing(WindowEvent event)
			{
			    dispose();	  // free the system resources
			    System.exit(0); // close the application
			}
		});
		this.setLayout(null);
		this.setSize(400,300);
		this.add(new CPanel());
	    }
	}
	
	new DriverFrame().show();
    }
    */
}
