/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 * Revision History:
 *   05/17/00:   release version 0.3 (initial version)
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;

/**
 * CShellMenuItem is the subclass of awt.MenuItem. It is the invisible 
 * beans and must be added to a CMenuButton widget. Unlike awt.MenuItem, 
 * the action event is already built into this widget. The user does not 
 * need to implement the ActionListener. 
 * <BR><BR>
 * CShellMenuItem is used to launch any non-java program. The name of this 
 * program must be set through setShellCommand (String cmd) method. When 
 * this menu item is selected, the built-in action is to disable this 
 * menuItem and start a new thread to execute the program. This thread will 
 * fork a process and wait until the program exits. It then re-enables the 
 * menu item. 
 * <BR><BR>
 * The default label for CShellMenuItem is String "[!]" . If user set the 
 * label, the "[!] "+label will be the real label to distinguish them
 * from awt.MenuItem. 
 *
 *
 * @version   0.3  06/24/00
 * @author    Ying Chen, Chip Watson
 *            
 * @since     adm0.3
 *
 */
public class CShellMenuItem extends MenuItem 
    implements ActionListener, Runnable {
    
    private String command = null;
    private String label = "";
    private Thread runner = null;

    /**
     * Null constructor. 
     */
    public CShellMenuItem () {
	this("");
    }

    /**
     * Constructor  with label parameter
     *
     * @param label    label string 
     */
    public CShellMenuItem (String label) {
	setLabel(label);
	addActionListener(this);
    }

   /**
     * Set the label string. The real label will be "[!] "+label.
     *
     * @param label      the label String 
     */
    public void setLabel (String label) {
	this.label = label;
	super.setLabel("[!] "+label);
    }


    /**
     * Return the label string (not include "[!] ").
     */
    public String getLabel() { return label; }
       
    /**
     * Set up the command string. 
     * The user must call this method.
     *
     * @param cmd     the command name need launched 
     */  
    public void setShellCommand (String cmd) { command = cmd; }

    /**
     * Return the command String.
     */
    public String getShellCommand () { return command; } 

    /**
     * Disable the menuItem and start a new thread
     */
    public void start () {
	setEnabled(false);
	if (runner == null || !runner.isAlive()) {
	    runner = new Thread(this);
	    runner.start();
	}
    }

    /**
     * execute and wait for the command. Print the exit code and
     * enable the menuItem after the command exit.
     */
    public void run () {
	try { 
	    int n = Runtime.getRuntime().exec(command).waitFor();
	    System.out.println("command exit code = "+n);
	}
	catch (Exception ioe) { 
	    System.out.println("Failed to exec "+command);
	}
	setEnabled(true);
    }

    public void actionPerformed (ActionEvent evt) {
	if (command == null) {
	    System.out.println("no command is set");
	    setEnabled(false);
	}
	else  { start();    }
    }

} // end of CShellMenuItem class
    
