/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *    04/17/00: Release version 0.2.
 *    08/28/00: Release version 0.4 (change all field to private).
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import cdev.*;
import cdev.data.*;

/**
 * CToggleButton implements a toggle button widget with 2 mutually 
 * exclusive buttons. This widget must be connected to a property of 
 * binary Enum type. The Enum strings will be the labels on the button. 
 * These two mutually exclusive buttons are represented by a raised 
 * and a etched rectangle. The etched button use to indicate the current 
 * state. Only the raised button works like a real push button. When the 
 * button is pressed the set action is sent to control system and the 
 * waiting cursor will appear. If the operation succeeds, the raised 
 * button will change to etched and etched button becomes raised. The 
 * cursor will revert to the default cursor. 
 * <BR><BR>
 * If the connected device/property is not of type of binary Enum, the 
 * "not" and "binary" strings will be the label of the buttons. No action 
 * perform when the either button is pressed. 
 * <BR><BR>
 * The toggle button can be stacked by ROW or by COLUMN using
 * setStacking(int) method. The default stacking is COLUMN. 
 *
 *
 * @version   0.4  09/08/00
 * @author    Ying Chen, Chip Watson 
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.2
 *
 */
public class CToggleButton extends CCanvas implements MouseListener {
    

    //--- for graphics paint ------------------------------------
    private int w, h, b0X, b0Y, b1X, b1Y = 0;
    private boolean b0R, b1R;
    private int s0X, s0Y, s1X, s1Y = 0;

    //--- boolean flags -----------------------------------------
    private boolean firstTime = true;   
    private boolean firstCall = true; 
    private boolean gotTags = false;
    private boolean eventLock = false;

    private static int VAL_TAG = Data.convertKey("value");
    private Data tdata = new Data();
    private DeviceListener setListener = null;
    private Cursor defCur  = new Cursor(Cursor.DEFAULT_CURSOR);
    private Cursor waitCur = new Cursor(Cursor.WAIT_CURSOR);
    private String[] tags = {"not", "binary"};
    private byte stacking = COLUMN;
    private byte value = 0;

    /**
     * Null constructor. 
     */
    public CToggleButton()  {
	addMouseListener(this);
    }
    
    /**
     * Null constructor  with device name and property name
     *
     * @param dName    device name
     * @param pName    property name
     */
    public CToggleButton (String dName, String pName)  {
	this();
	setDeviceName(dName);
	setPropertyName(pName);
    }

    /**
     * Set the stacking position
     * @param newStacking   new stacking position, which may be one of:
     *             <OL>
     *             <LI>  CConstants.COLUMN (default)
     *             <LI>  CConstants.ROW
     *             </OL>
     */
    public void setStacking (byte newStacking) {
	stacking = newStacking;
    }

    /**
     * Get the stacking position
     */
    public byte getStacking () { return stacking; }


    private void initGraphics () {
	Dimension rec = getSize();
	FontMetrics fm = Toolkit.getDefaultToolkit().getFontMetrics(getFont());
	int s0W = fm.stringWidth(tags[0]);
	int s1W = fm.stringWidth(tags[1]);
	int fontH = fm.getHeight() - fm.getDescent();
	if (stacking == ROW) {
	    w = rec.width;
	    h = (int) (rec.height/2);
	    b1X = 0;
	    b1Y = h;
	    s0X = (int) ((w - s0W)/2);
	    s0Y = h - (int)((h - fontH)/2);
	    s1X = (int) ((w- s1W)/2); 
	    s1Y = h + s0Y;
	}
	else {
	    w = (int) (rec.width/2);
	    h = rec.height;
	    b1X = w;
	    b1Y = 0;
	    s0X = (int) ((w - s0W)/2);
	    s0Y = h - (int)((h - fontH)/2);
	    s1X = w + (int) ((w - s1W)/2);
	    s1Y = s0Y;		
	}
    }   

    /**
     * Draw two buttons and labels
     */
    public void paint(Graphics g) {
	if (!gotTags) return;
	if (firstTime) {
	    initGraphics();
	    firstTime = false;
	}
	//--- draw the bouutons -----------------------
	if (value == 0) 
	    { b0R = false;   b1R = true; }
	else    
	    { b0R = true;   b1R = false; }
	g.setColor(getBackground());
	g.fill3DRect(b0X, b0Y, w, h, b0R);
	g.fill3DRect(b1X, b1Y, w, h, b1R);

	//--- draw the labels on the buttons ---------
	g.setColor(getForeground());
	g.drawString(tags[0], s0X, s0Y);
	g.drawString(tags[1], s1X, s1Y);
    }


    private byte getPressedButton (int x, int y) {
	byte buttonNumber = 0;
	if (stacking == ROW && y > h)       buttonNumber = 1;
	if (stacking == COLUMN && x > w)    buttonNumber = 1;
	return buttonNumber;
    }

    /**
     * Handle the mouse press. Set the evnetLock to true only 
     * when the up-button is pressed. 
     */
    public void mousePressed (MouseEvent me) {
	eventLock = false;
	if(!tags[0].equals("none") && !tags[1].equals("ENUM")) { 
	    int x = me.getX();
	    int y = me.getY();
	    byte pressButton = getPressedButton(x, y);
	    //--- only handle the raise button ----------
	    if (pressButton != value)  	eventLock = true;
	}
    }


    /**
     * To send the new value to the control system. The cursor change to
     * waiting cursor. The setListener will keep track the error message
     * return from the control system and set the cursor back to default
     * cursor.
     */     
    public void mouseReleased (MouseEvent me) { 
	if (!eventLock)  return;
	//--- setListener to check the set command -------------
	if (setListener==null) setListener = new DeviceListener() {
	    public void deviceChanged(DeviceEvent evt) {
		//--- update the value and repaint the buttons
		if (evt.getError() == null) {
		    System.out.println("success to set value ("+value+")");
		    value = (byte)(value^0x1);
		    repaint();
		} else {
		    System.out.println("failed to set value");
		}
		setCursor(defCur);
	    }
	};
	//--- action is change the cursor, send the set command ---
	byte val = (byte)(value^0x1);
	tdata.insert(VAL_TAG, String.valueOf(val));
	try {
	    //System.out.println(tdata);
	    setCursor(waitCur);
	    device.set (propertyName, tdata, setListener);
	    device.getControlSystem().flush();
	} catch (Exception ex) {
	    System.out.println ("failed to send value ("+val+")");
	}
    }	
    

    public void mouseClicked (MouseEvent me) { }
    public void mouseEntered (MouseEvent me) { }
    public void mouseExited (MouseEvent me) { }

    /**
     * Get the Enum value from control system and add to the choice
     * at first call. Change the select item to the current Enum
     * when the control system updates.
     */ 
    public void deviceChanged(DeviceEvent event) {
	if (firstCall) {
	    try {
		DataEntry de = event.getData().getDataEntry("value");
		if (de.getValue() instanceof String[]) {
		    //--- need to check the size ??? here ----
		    tags = de.stringArray();
		    System.out.println("devCh:: current state = "+tags[0]);
		}
		else { System.out.println("none binary type"); }
		gotTags = true;
		startMonitor();
		firstCall = false;
	    }
	    catch (Exception e) { e.printStackTrace(); }   
	}
	else {
	    try {
		String s = event.getData().getDataEntry("value").stringValue();
		value = getTagIndex(s);
		//try { value = Byte.parseByte(s); }
		//catch(NumberFormatException ne) { value = getTagIndex(s); }
	    }
	    catch (Exception e) { e.printStackTrace(); }  
	    repaint();
	}
    }
    
    private byte getTagIndex(String tagName) {
	byte index = 0;
	boolean found = false;
	for (index=0; index<tags.length && !found; index++) {
	    if (tags[index].equals(tagName)) 
		found = true;
	}
	return (byte)(index-1);
    }

}  // end of CToggleButton










