/*
 * Copyright (c) 1996,1997 Southeastern Universities Research Association,
 *                         Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 * Revision History:
 *    8/07/00 : Release version 0.4. (Initial version)
 *
 */

package	adm;

import java.awt.*;
import java.util.Vector;
import java.io.*;

import adm.event.*;
import cdev.*;
import cdev.data.*;

/**
 * CTrace is a data model for CCartesin widget. It connects to two 
 * device/properties with the simultaneous data update. One is the value 
 * of x-axis and another is the y-axis. The time stamp is used to trig the 
 * trace event. This means only the new data pair with same time stamp will 
 * fire up CTraceEvent (send a trace event to Listeners).
 * <BR><BR>
 * CTrace class defines many trace properties. You may specify one of the 
 * three axes range styles (x-axis or y-axis), RANGE_AUTO, RANGE_STATIC,
 * RANGE_FROM_PROPRTY with setRangeStyle. The default range style for all 
 * axes are RANGE_FROM_PROPRTY. Call setTraceColor method to reset the default 
 * trace color (green) and use setTraceStyle to select STYLE_POINT or STYLE_LINE 
 * (default) as the trace style. 
 * <BR><BR>
 * The setX(Y)DisplayLo(Hi) methods to specify the display range. When any of 
 * these methods called, the range style for that axis will be forced to 
 * RANGE_STATIC. If the range style is set to RANGE_STATIC and range low 
 * and/or range high are not set, the default display range (0 for low and 
 * 100 for high) will be used. 
 *
 *
 * @version   0.4  09/17/00
 * @author    Chip Watson, Ying Chen 
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.4
 *
 */
public class CTrace extends Component implements CConstants {
 
    static final String BLANK = "";
    static final double BAD_VALUE = -11111;

    static DeviceContext defContext;
    static { 
        Data defContextData = new Data();
	defContextData.insert("time", 1);
        defContext = new DeviceContext(defContextData);
    }
    static DeviceContext limitsContext;
    static {
        Data limitsContextData = new Data();
        limitsContextData.insert("displayHigh", 1);
	limitsContextData.insert("displayLow", 1);
	limitsContextData.insert("time", 1);
        limitsContext = new DeviceContext(limitsContextData);
    }

    private Device[] devices = new Device[2];
    private String[] deviceNames = new String[2];
    private String[] propertyNames =  new String[2];
    private String[] lastDeviceNames =  new String[2];
    private String[] lastPropertyNames =  new String[2];
    private AListener[] listeners = {new AListener(0), new AListener(1)};

    private double[] displayHi = {100, 100};
    private double[] displayLo = {0, 0};
    private double[] timestamp = {BAD_VALUE, BAD_VALUE};  
    private boolean[] firstCall = {true, true};
    private boolean[] isArrayData = {false, false};
    private int[] rangeStyle = {RANGE_FROM_PROPERTY, RANGE_FROM_PROPERTY};
    private Vector tListeners = new Vector();

    private Color traceColor = Color.black;
    private int traceStyle = STYLE_LINE;
    private Axis xAxis = new Axis(CConstants.AXIS_BOTTOM);
    private Axis yAxis = new Axis(CConstants.AXIS_LEFT);
    public boolean redraw = false;
    private double[] valX = new double[1];
    private double[] valY = new double[1];

    /**
     * Null constructor. 
     */
    public CTrace () { }

    /**
     * Constructor with parameters
     *
     * @param xDName     the name of device for x value
     * @param xPName     the name of property for x value
     * @param yDName     the name of device for y value
     * @param yDName     the name of property fot y value
     */
    public CTrace (String xDName, String xPName, String yDName, String yPName) {
	setXDeviceName(xDName);
	setXPropertyName(xPName);
	setYDeviceName(yDName);
	setYPropertyName(yPName);
    }

    /**
     * Set up the X axis of this trace
     * @params axis   the new x axis
     */ 
    public void setXAxis (Axis axis) { xAxis = axis; }

    /**
     * Return the X axis of this trace
     */
    public Axis getXAxis () { return xAxis; }

    /**
     * Set up the Y axis of this trace
     * @params axis   the new y axis
     */ 
    public void setYAxis (Axis axis) { yAxis = axis; }

    /**
     * Return the Y axis of this trace
     */
    public Axis getYAxis () { return yAxis; }
   

    /**
     * Add the data listener to receive the notification
     */
    public synchronized void addTraceListener (CTraceListener tl) {
        if (!tListeners.contains(tl))
            tListeners.addElement(tl);
    }

    /**
     * Remove the data listener service
     */
    public synchronized void removeTraceListener (CTraceListener tl) { 
        if (tListeners.contains(tl))
            tListeners.removeElement(tl);
    }

    /**
     * Fire the event change - notify all the data listener
     */
    public void fireTraceDataChange () {
        CTraceEvent de = new CTraceEvent(this, valX, valY);
        //--- make a copy of the listener vector ---------------
        //--- in case it changes while we firing events --------
        Vector v;
        synchronized (this) { v = (Vector)tListeners.clone(); }

        //--- fire the event to all listeners -----------------
        for (int i=0; i< v.size(); i++ ) 
            ((CTraceListener) v.elementAt(i)).receivedNewData(de);
    }

    /**
     * Set limitsContext to get all meta-data at first time
     */
    synchronized protected void getFirstValue (int i) 
    {
	try {
	    if (deviceNames[i]==null || propertyNames[i] == null)   return;
	    if (!deviceNames[i].equals(lastDeviceNames[i])) {
		devices[i] = new Device(deviceNames[i]);
		lastDeviceNames[i] = deviceNames[i];
	    }
	    devices[i].setContext (limitsContext);
	    devices[i].get(propertyNames[i], listeners[i]);
	    firstCall[i] = true; 
	}
	catch(Exception e) { e.printStackTrace(); }
    }
    
    
    /**
     * Set the Device name 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setXDeviceName(String newName) 
    {
	newName = CUtility.makeValidName(newName);
	if (!CUtility.testNameMatch(deviceNames[0], newName)) {
	    stopMonitor(0);
	    deviceNames[0] = newName;
	    getFirstValue (0);
	} 
    }
    
    /**
     * Return the device name to which this component is connected.
     * If device name is not set, return empty string "". 
     */
    public String getXDeviceName()
    {
	return (deviceNames[0]==null)?BLANK:deviceNames[0];
    }
    
    /**
     * Set device property name 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setXPropertyName(String newName)
    {
        newName = CUtility.makeValidName(newName);
        if (!CUtility.testNameMatch(propertyNames[0], newName)) {
	    stopMonitor(0);
	    propertyNames[0] = newName; 
	    getFirstValue (0);        
        }
    }
    
   /**
     * Return the device property name to which this component is connected.
     * If property name is not set, return empty string "". 
     */
    public String getXPropertyName()
    {
        return (propertyNames[0]==null)?BLANK:propertyNames[0];
    }

    /**
     * Set the Device name 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setYDeviceName(String newName) 
    {
	newName = CUtility.makeValidName(newName);
	if (!CUtility.testNameMatch(deviceNames[1], newName)) {
	    stopMonitor(1);
	    deviceNames[1] = newName;
	    getFirstValue (1);
	} 
    }
    
    /**
     * Return the device name to which this component is connected.
     * If device name is not set, return empty string "". 
     */
    public String getYDeviceName()
    {
	return (deviceNames[1]==null)?BLANK:deviceNames[1];
    }
    
    /**
     * Set device property name 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setYPropertyName(String newName)
    {
        newName = CUtility.makeValidName(newName);
        if (!CUtility.testNameMatch(propertyNames[1], newName)) {
	    stopMonitor(1);
	    propertyNames[1] = newName; 
	    getFirstValue (1);        
        }
    }

   /**
     * Return the device property name to which this component is connected.
     * If property name is not set, return empty string "". 
     */
    public String getYPropertyName()
    {
        return (propertyNames[1]==null)?BLANK:propertyNames[1];
    }

    protected void stopMonitor(int i) 
    {
	if (listeners[i] == null) return;
        if (listeners[i].active) {
	    try {
		listeners[i].active = false;
		devices[i].monitorOff(propertyNames[i], listeners[i]);
	    } 
	    catch (IOException e) {
		e.printStackTrace();
	    }
	}
    }
    
    protected void startMonitor(int i) 
    {
	try {
	    if (listeners[i].active) stopMonitor(i);
	    if (deviceNames[i] == null || propertyNames[i] == null) return;
	    if (!deviceNames[i].equals(lastDeviceNames[i])) {
		devices[i] = new Device(deviceNames[i]);
		lastDeviceNames[i] = deviceNames[i];
	    }    
	    devices[i].setContext (defContext);
	    devices[i].monitorOn(propertyNames[i], listeners[i]);
	    lastPropertyNames[i] = propertyNames[i];
	    listeners[i].active = true;
	} 
	catch (IOException e) {
	    listeners[i].active = false;
	    e.printStackTrace();
	}
    }
    
    /*
     * for CCartesian widget to use
     */
    protected boolean gotRange() {
	return (!firstCall[0] && !firstCall[1]);
    }

    /**
     * Set the color of this trace
     */
    public void setTraceColor(Color aColor) { 
	traceColor = aColor;
	yAxis.setLabelColor(aColor);
    }
    
    /**
     * Return the color of this trace
     */
    public Color getTraceColor () {
	return traceColor;
    }
    
    /**
     * Return the trace style of this trace
     */
    public int getTraceStyle () { return traceStyle; }

    /**
     * Set up the trace style of this trace
     * @params style    the new trace style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.STYLE_LINE
     *             <LI>  CConstants.STYLE_POINT
     *             </OL>
     */
    public void setTraceStyle (int style) { traceStyle = style; }
 
    /**
     * Return the X axis range style of this trace.
     */
    public int getXRangeStyle () { return rangeStyle[0]; }

    /**
     * Set up the X axis range style of this trace.
     * @params style    the X axis range style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.RANGE_AUTO
     *             <LI>  CConstants.RANGE_STATIC
     *             <LI>  CConstants.RANGE_FROM_PROPERTY
     *             </OL>
     */
    public void setXRangeStyle (int style) { rangeStyle[0] = style; }

    /**
     * Return the Y axis range style of this trace.
     */
    public int getYRangeStyle () { return rangeStyle[1]; }

    /**
     * Set up the Y axis range style of this trace.
     * @params style    the X axis range style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.RANGE_AUTO
     *             <LI>  CConstants.RANGE_STATIC
     *             <LI>  CConstants.RANGE_FROM_PROPERTY
     *             </OL>
     */
    public void setYRangeStyle (int style) { rangeStyle[1] = style; }

    /**
     * Set up the Y axis position of this trace
     * @param position  The Y axis position, which may be one of:
     *                  <ol>
     *                     <li>CConstants.AXIX_LEFT
     *                     <li>CConstants.AXIS_RIGHT
     *                  </ol>
     */
    public void setYAxisPosition (int position) {
	if (position == AXIS_LEFT || position == AXIS_RIGHT)
	    yAxis.setPosition (position);
    }

    /** 
     * Set X axis display range (low and high). 
     *
     * @param low      the x-axis display low value 
     * @param high     the x-axis display high value
     */
    public void setXDisplayRange (double low, double high)
    {
	rangeStyle[0] = RANGE_STATIC;
	displayLo[0] = Math.min(low, high);
	displayHi[0] = Math.max(low, high);
    }

    /** 
     * Set Y axis display range (low and high). 
     *
     * @param low     the y-axis display low value 
     * @param high     the y-axis display high value
     */
    public void setYDisplayRange (double low, double high)
    {
	rangeStyle[1] = RANGE_STATIC;
	displayLo[1] = Math.min(low, high);
	displayHi[1] = Math.max(low, high);
    }

    /**
     * set x-axis display high.
     *
     * @param high     the x-axis display high value
     */
    public void setXDisplayHi (double high) { 
	rangeStyle[0] = RANGE_STATIC;
	displayHi[0] = high; 
    }

    /**
     * set y-axis display high.
     *
     * @param high     the y-axis display high value
     */
    public void setYDisplayHi (double high) {
	rangeStyle[1] = RANGE_STATIC;
	displayHi[1] = high; 
    }

    /**
     * set x-axis display low.
     *
     * @param low     the x-axis display low value
     */
    public void setXDisplayLo (double low) { 
	rangeStyle[0] = RANGE_STATIC;
	displayLo[0] = low; 
    }

    /**
     * set y-axis display low.
     *
     * @param low     the y-axis display low value
     */
    public void setYDisplayLo (double low) { 
	rangeStyle[1] = RANGE_STATIC;
	displayLo[1] = low; 
    }

    /**
     * Return x-axis display high
     */
    public double getXDisplayHi () { return displayHi[0]; }
    
    /**
     * Return y-axis display high
     */
    public double getYDisplayHi () { return displayHi[1]; }

    /**
     * Return x-axis display low
     */
    public double getXDisplayLo () { return displayLo[0]; }
    
    /**
     * Return y-axis display low
     */
    public double getYDisplayLo () { return displayLo[1]; }

    
    //*********************************************************************
    
    /**
     * This class implements deviceListener which has a fuction
     * deviceChanged(). This gets the data from the system.
     */    
    private class AListener implements DeviceListener {
	    
	private int which;
	private boolean active = false;

	/**
	 * constructor
	 * @param i    i = 0/1
	 */
	AListener(int i) {
	    if (which != 0 && which != 1) return;
	    which = i;
	}
	
	/**
	 * method inhereted from DeviceListener
	 * @param event  A DeviceEvent
	 */
	public void deviceChanged(DeviceEvent de) {
	    double[] pretty;
	    DataEntry dEntry = de.getData().getDataEntry("value");
	    if (firstCall[which]) {
		double low, high;
		try {
                    if (dEntry.getDimensions() != null) { 
                        isArrayData[which] = true;
                        if (which == 0)   valX = dEntry.doubleArray();
			else              valY = dEntry.doubleArray();
                        System.out.println("get array data first time");
                    }
                    else {   // data is not array
                        if (which == 0)   valX[0] = dEntry.doubleValue();
			else              valY[0] = dEntry.doubleValue();
                    }
		    switch (rangeStyle[which]) {
		    case RANGE_STATIC:
			low = displayLo[which];
			high = displayHi[which];
			break;
		    case RANGE_AUTO: 
			double[] val;
			if (which == 0)   val = valX;
			else              val = valY;
			if (isArrayData[which]) {
			    low = CUtility.arrayMin(val);
			    high = CUtility.arrayMax(val);
			}
			else {
			    low = val[0] - 0.01*Math.abs(val[0]);
			    high = val[0] + 0.01*Math.abs(val[0]);
			}
			break;
		    default:    //case RANGE_FROM_PROPERTY
			low = de.getData().getDataEntry("displayLow").doubleValue();
			high = de.getData().getDataEntry("displayHigh").doubleValue();
		    }
		    pretty = CUtility.makePrettyAxis(low, high);
		    displayLo[which] = pretty[0];
		    displayHi[which] = pretty[1];
		    if (which == 0) xAxis.setMinorTick((int)pretty[2]);
		    else            yAxis.setMinorTick((int)pretty[2]);
		    firstCall[which] = false;
		    startMonitor(which);
		} catch (Exception e) {
		    //e.printStackTrace();
		    System.out.println("failed to get data at first time");
		}
	    }
	    else {
		try {
		    if (isArrayData[which]) {
			if (which == 0)   valX = dEntry.doubleArray();
			else              valY = dEntry.doubleArray();
		    }
		    else {   // data is not array
			if (which == 0)    valX[0] = dEntry.doubleValue();
			else               valY[0] = dEntry.doubleValue();
		    }
		    timestamp[which] = de.getData().getDataEntry("time").doubleValue();
		    if (rangeStyle[which] == RANGE_AUTO) do_auto(which);
		    if (timestamp[0] == timestamp[1])  fireTraceDataChange();
		} catch (Exception e) {
		    System.out.println("failed to get the new data");
		}
	    }
	}

    private void do_auto (int j) {
	boolean rangeChanged = false;
	redraw = false;
	double[] val;
	if (j == 0)  val = valX;
	else         val = valY;
	double min = displayLo[j];             
	double max = displayHi[j];  
        if (max < CUtility.arrayMax(val)) {
	    rangeChanged = true;
	    max = CUtility.arrayMax(val);
	}
	if (min > CUtility.arrayMin(val)) {
	    rangeChanged = true; 
	    min = CUtility.arrayMin(val);
	}
        if (rangeChanged) {
            double[] pretty = CUtility.makePrettyAxis(min, max);
	    displayLo[j] = pretty[0];
	    displayHi[j] = pretty[1];
	    if (j == 0)   xAxis.setMinorTick((int)pretty[2]);
	    else          yAxis.setMinorTick((int)pretty[2]);
	    //System.out.println("CTrace:: display "+j+" lo="+displayLo[j]
	    //	       +", hi="+displayHi[j]);
	    redraw = true;
	    rangeChanged = false; 
	    fireTraceDataChange();
        }
    }


    }
}












