/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  03/10/00:  Release version 0.1.
 *  04/20/00:  Release 0.2 (new format method used format class).
 *  05/08/00:  Release 0.3 (makePrettyAxis method).
 *  08/07/00:  Release 0.4 (changed all methods to public).
 *  08/23/00:  Release 0.4 (add arrayMax and arrayMin methods).
 *
 */

package adm;

import java.text.NumberFormat;
import cdev.ControlSystem;

/**
 * This class contains the utilities used by the adm package.
 *
 *
 * @version   0.4  09/17/00
 * @author    Ying Chen, Chip Watson
 *           
 * @since     adm0.1
 *
 */
public class CUtility {

    static final double BASE = 10;
    static final double SMALL = 0.00001;

    private CUtility() { }
    
    public static String makeValidName(String name)
    {
        if (name != null && name.length()==0)
	    name = null; // treat empty string as null
	return name;
    }
    
    public static boolean testNameMatch(String oldName, String newName)
    {
        boolean match = false;
        if (newName == null) {
	    if (oldName == null)
		match = true;
	}
	else if (newName.equals(oldName))
	    match = true;
	return match;
    }
    
    
    public static void connect(String host, int port) {
        try {
            ControlSystem cs = ControlSystem.getDefault();
            cs.connect(host,port);
        } catch (Exception e) {
            System.out.println("failed to connect to "+host+" "+port);
        }
    }
    
    
    public static String format(double x) {
	NumberFormat numberFormat = NumberFormat.getNumberInstance();
	numberFormat.setMaximumFractionDigits(3);
	return numberFormat.format(x);
    }
    
    
    public static String format(double x, String fmt) {
	Format format;
	if (fmt == null || fmt == "") 
	    format = new Format("%.3g");
	else 
	    format = new Format(fmt);
	String returnString = format.form(x);
	char code = format.getFormatCode();
	if (code=='d' || code=='i' || code=='o' || code=='x' || code=='X') {
	    long y = new Double(x).longValue();
	    returnString = format.form(y);
	}
	return returnString;
    }
    
    /**
     * Generate pretty axis and tick number.
     * the tick number will be 6 - 10 depend on the range
     */
    public static double[] makePrettyAxis (double low, double high) {
	double[] lo_hi_tick = new double[3];
	//--- in case low and high are same -----------
	double diff = high - low;
	if (Math.abs(diff) < SMALL) {
	    if (high != 0)  high += 0.1*Math.abs(high);
	    else            high += 0.1;
	}
	double k = Math.floor(Math.log(diff)/Math.log(BASE));
	double tenKpow = Math.pow(BASE, k);
	lo_hi_tick[0] = tenKpow*Math.floor(low/tenKpow);
	lo_hi_tick[1] = tenKpow*Math.ceil(high/tenKpow);
	int tick = (int)((lo_hi_tick[1]-lo_hi_tick[0])/tenKpow);
	if (tick <=2 || tick == 5)   tick = 10;
	if (tick == 3 || tick == 4)  tick *= 2;
	lo_hi_tick[2] = tick;
	return lo_hi_tick;
    }
    
    public static double arrayMax (double[] array) {
	double max = array[0];
	for (int i = 1; i<array.length; i++) {
	    if (array[i]-max > 0) max = array[i];
	}
	return max;
    }
    
    public static double arrayMin (double[] array) {
	double min = array[0];
	for (int i = 1; i<array.length; i++) {
	    if (array[i]-min < 0) min = array[i];
	}
	return min;
    } 
   
}
