//---------------------------------------------------------------------------
// Copyright (c) 1999 Southeastern Universities Research Association,
//                    Thomas Jefferson National Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description:
//      Base class for creating CDEV server applications.
//      Derived classes must supply protocol handlers for client sockets
//
//  Author: Chip Watson
//
//----------------------------------------------------------------------------
// Revision History:
//
// 26-Apr-99 : CW Initial version.
// 26-Sep-99 : CW Added flush prior to sleep
//  9-Apr-00 : CW Rearranged threads for better symmetry, slowed context switches
//

package cdev;
import cdev.io.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *
 * @version Beta2-1
 * <br>
 * Server class which manages CDEV transactions for multiple clients.
*/
public abstract class BasicServer extends Thread
  implements TransactionListener, TransactionResponseListener
{
  static final int BUFFER_BEFORE_WAKE=10; // buffer 10 transactions

  ServerSocket server;
  TransactionManager tmgr;
  Fifo inbound;
  Fifo outbound;

  Thread inWorker;
  Thread outWorker;

  Vector services = new Vector();

  public BasicServer () {
    tmgr = new TransactionManager();
  }

  /**
   * Initialize the server, starting 3 threads:
   *   1 --> listen for connections (new thread for each)
   *   2 --> process queue for inbound messages
   *   3 --> process queue for outbound responses
   */
  public int initialize(int port) throws IOException {

    inbound = new Fifo(); // inbound work
    outbound = new Fifo(); // outbound responses

    // thread to process inbound queue
    inWorker = new Thread() {
      public void run () {
	processInbound();
      }
    };
    inWorker.setDaemon(true);
    inWorker.start();

    // thread to process outbound queue
    outWorker = new Thread() {
      public void run () {
	processOutbound();
      }
    };
    outWorker.setDaemon(true);
    outWorker.start();

    // thread to accept connections
    server = new ServerSocket(port);
    this.setDaemon(true);
    this.start();

    return server.getLocalPort();
  }

  public int initialize() throws IOException {
    int port = 0;
    return initialize(port);
  }

  /** 
   * Run method accepts new connections and starts a client
   * handler (method supplied in derived class)
   */
  public void run () {
    System.out.println("Server listening on port "+server.getLocalPort());
    while (true) {
      try {
	Socket client = server.accept();
	System.out.println("received a connection from "+client);
	startClient(client);
      } catch (Exception e) {
	System.out.println("Failed to start client handler "+e);
      }
    }
  }

  protected abstract void startClient(Socket client) throws IOException ;
  protected abstract void process (DeviceTransaction dt);

  /** 
   * Post a new transaction for processing in server (coming from
   * a client handler).
   *
   */
  public void post(DeviceTransaction dt) {

    //System.out.println("BasicServer post..."+dt);
    tmgr.add(dt); // list all active transactions w/ manager
    inbound.add (dt); // add to inbound queue

    // could avoid following sync and notify if IOService had a way
    // of indicating it had no more to post...
    try {
      synchronized (inWorker) {
	inWorker.notify(); // Notify worker thread
      }
    } catch (Exception e) {e.printStackTrace();}
  }

  /** 
   * Cancel a transaction
   *
   */
  public void cancel(DeviceTransaction dt) {}

  /** 
   * Get an enumeration of transactions
   *
   */
  public Enumeration getTransactions() {
    return tmgr.elements();
  }

  /** 
   * Queue a completed transaction for transmission
   * (coming from derived class's process routine)
   *
   */
  public void handleResponse (DeviceTransaction dt) {
    //tmgr.process (dt);
    outbound.add (dt);
    if (outbound.size() > BUFFER_BEFORE_WAKE || inbound.isEmpty()) {
      synchronized (outWorker) {
	outWorker.notify();  // Notify finish thread
      }
    }
  }


  public void processOutbound () {
    while (true) {
      //--- wait to be notified by arrival of work ---
      synchronized (outWorker) {
	if (outbound.isEmpty())
	  try { outWorker.wait(); } catch (InterruptedException e) {}
      }
      //--- process all transactions in the outbound fifo ---
      //System.out.println("process outbound...");
      tmgr.process (outbound);
      flush(); // improve responsiveness for command-response
    }
  }

  private void processInbound() {
    while (true) {
      //--- wait to be notified by the service ---
      //System.out.println("processInBound...");
      synchronized (inWorker) {
	if (inbound.isEmpty()) {
	  try { inWorker.wait(); } catch (InterruptedException e) {}
	}
      }
      //System.out.println("worker awake...");
      while (inbound.notEmpty()) {
	DeviceTransaction dt = (DeviceTransaction) inbound.next();
	// System.out.println("processInBound "+dt);
	if (dt.getState() != DeviceTransaction.CANCELLED) {
	  try {
	    process(dt);
	  } catch (Exception e) {
	    System.out.println("BasicServer caught process exception\n"+e);
	  }
	}
      }
    }
  }

  protected boolean more() {return inbound.notEmpty();}
  protected DeviceTransaction next() { return (DeviceTransaction) inbound.next();}

  //************************************************************************//
  /**
   * Register an IOService with this Server
   */
  synchronized public void registerIOService(IOService s) {
    services.removeElement(s);
    services.addElement(s);
  } 
  synchronized public void unregisterIOService(IOService s) {
    services.removeElement(s);
    Enumeration trans = getTransactions();
    while (trans.hasMoreElements()) {
      DeviceTransaction dt = (DeviceTransaction)trans.nextElement();
      if (dt.listener == s) {
	cancel(dt);
	tmgr.removeTrans(dt);
	dt.state = DeviceTransaction.CANCELLED;
      }
    }
  }

  //************************************************************************//
  /**
   * Flushes any pending I/O .
   */
  void flush ()
  {
    for (int i=0;i<services.size();i++) {
      IOService service = (IOService) services.elementAt(i);
      try {
	service.flush();
      } catch (IOException e) {
	System.out.println("flush error on "+service);
	e.printStackTrace();
      }
    }
  }

}
