// $Id: DeviceArray.java,v 1.2 1999/07/20 18:58:07 watson Exp $
// (c) Copyright ...
//----------------------------------------------------------------------------
//
//  Description:
//      Device class implementing collection of devices
//
//  Authors: Franck Di Maio
//
//----------------------------------------------------------------------------

package cdev;

import cdev.data.*;
import cdev.dir.*;

import java.io.*;		// for IOException

/**
 * This class provides an ordered collection of devices represented by a Device object.
 * It overwrites the I/O methods.
 * The data objects referred in I/O methods and in DeviceEvent include
 * the values for all the single devices.
 * As a result, the "value" dataEntry will have as its first dimension
 * the size of the collection.
 * @author Franck Di Maio
 * @version $Revision: 1.2 $
 * @since JDK1.1
 */

public class DeviceArray extends Device {

  //**************************************************************************
  // Data
  //**************************************************************************
  
  /* The deviceName (inherited) holds an user-specified name that is either 
   * an explicit parameter of the constructor or the name of a pre-defined list. */

  /* The parentSys and context (inherited) have the same usage as any device. */

  /** The size hold the number of elements.
   *  It is initialized by the constructor.
   */
  private int size;

  /** The deviceList holds the directory service list when this device array 
   *  is created from a named list. */
  private DeviceList deviceList;

  /** The deviceData array holds each component device description. */
  private DeviceData[] deviceDataArray;

  /** The compNames holds the names of the components. */ 
  private String[] compNames;

  //**************************************************************************
  // Constructors
  //**************************************************************************

    /** 
     * Constructs a device array from a pre-defined list.
     * The list must be defined into the directory service.
     * @param listName a name of a list of devices.
     * @param controlSystem the control system that will manage the I/O for this device,
     *                      null means the default control system.
     * @exception DataNotFoundException if the list id not found in the directory service.
     */
    public DeviceArray (String listName, ControlSystem controlSystem) 
	throws DataNotFoundException , IOException{
	super (listName, controlSystem);
	DirectoryService ds;
	try {
	    ds = Config.getDirectoryService();
	}
	catch (Exception e) {
	    throw new DataNotFoundException ("Cannot find the list " + listName
					     + ": " + e);
	}
	// get an array of DeviceData from the directory service
	deviceList =  ds.getDeviceListByName(listName);
    }

    /**
     * Constructs a device array from a pre-defined list.
     * Use the default control system.
     * Doc to be completed.
     * @see DeviceArray(String,ControlSystem)
     */
    public DeviceArray (String listName) 
	throws DataNotFoundException, IOException {
	this(listName, (ControlSystem) null);
    }

    /** 
     * Constructs a device array from a dynamic list.
     * @param name the name for the array device
     * @param names the names of the devices that compose the array.
     * @param controlSystem the control system that will manage the I/O for this device,
     *                      null means the default control system.
     * @exception DataNotFoundException if no device data found for one componant device.
     */
    public DeviceArray (String name, String[] names, ControlSystem controlSystem) 
      throws DataNotFoundException, IOException {
	super (name, controlSystem);
	if (names.length == 0)
	    throw new IllegalArgumentException (name + " empty array of device names");
	size = names.length;
	compNames = names;	// ? should clone the array
    }

    /** 
     * Constructs a device array from a dynamic list.
     * Use the default control system.
     * Doc to be completed.
     * @see DeviceArray(String,String[],ControlSystem)
     */
    public DeviceArray (String name, String[] names) 
      throws DataNotFoundException , IOException{
	this(name,names,null);
    }

  //**************************************************************************
  // overwritten methods
  //**************************************************************************

  /**
   * Returns the reference to the device class this device belongs to.
   * A unique common device class must be found among the components.
   * @exception DataNotFoundException if no device class found or if more than one.
   */
  public DeviceClass getDeviceClass() throws DataNotFoundException {
      DeviceData[] dds = getDeviceDatas();
      DeviceClass dc = dds[0].getDeviceClass(); // class of the 1st device
      for (int i=1; i < size; i++)
	  if (!dc.equals(dds[i].getDeviceClass())) // a different class found
	      throw new DataNotFoundException(getName() + " components of different classes: "
					      + dc.getName()  + " and "
					      + dds[i].getDeviceClass().getName());
      return dc;
  }

  /**
   * Returns true if the I/O may require a cycle type.
   * Returns true if any of the component devices need a cycle type,
   * false otherwise.
   */
    public boolean needCycleType() throws DataNotFoundException {
	DeviceData[] dds = getDeviceDatas();
	for (int i=0; i<size; i++)
	    if (dds[i].needCycleType())
		return true;
	return false;
    }

    /* needCycleType(String propertyName) is not overwritten */

  //**************************************************************************
  // additional methods
  //**************************************************************************
  
  /**
   * Returns the number of components in this device array.
   */
  public final int size() {
      return size;
  }

  /** Returns the array of DeviceData */
  public DeviceData[] getDeviceDatas() 
      throws DataNotFoundException 
    {

      if (deviceDataArray != null)
	  return deviceDataArray;

      // initialize deviceDataArray depending on construction options
      if (deviceList != null) {	// this DeviceArray was created from a named list
	  deviceDataArray = deviceList.getDeviceData();
	  return deviceDataArray;
      }
      if (compNames != null) {	// this deviceArray was created from an array of names
	  DirectoryService ds;
	  try {
	      ds = Config.getDirectoryService();
	  }
	  catch (Exception e) {
	      throw new DataNotFoundException ("Cannot get the components's DeviceData of " + getName()
					       + ": " + e);
	  }
	  deviceDataArray = new DeviceData[size];
	  for (int i=0; i<size; i++)
	      deviceDataArray[i] = ds.getDeviceData(compNames[i]);
	  return deviceDataArray;
      }
      // shouldn't fall into this throw
      throw new DataNotFoundException ("Cannot find DeviceData for components of " + getName());

  }

  /**
   * Returns the names of the components in this device array. */
  public String[] getNames() 
      throws DataNotFoundException
    {
      if (compNames == null) { // this DeviceArray wasn't created from a list of names
	  DeviceData [] dds = getDeviceDatas();
	  compNames = new String[size];
	  for (int i=0; i<size; i++)
	      compNames[i] = dds[i].getName();
      }
      return compNames;
    }


  private Device[] components;
  /**
   * Returns component as independant devices.
   * Returns an array of devices sharing the same context and the
   * same control system as this device.
   */
  public final Device[] getDevices() 
      throws DataNotFoundException, IOException
    {
      if (components == null) {
	  components = new Device[size];
	  ControlSystem cs = this.getControlSystem();
	  Device comp;
	  DeviceData[] dds = getDeviceDatas();
	  for (int i=0; i<size; i++) {
	      comp = new Device(dds[i].getName(), cs); // shall be replaced with a new Device(dds[i]...
	      comp.setContext(this.getContext());
	      components[i] = comp;
	  }
      }
      return components;
    }
}
