//-----------------------------------------------------------------------------
//
//  Description:
//      Context class for Device
//
//  Author: Chip Watson
//
//-----------------------------------------------------------------------------
// Revision History:
//
// 28-Aug-98  CW Prototype version
// 24-Sep-98  CW added accessor for the edit version number, getSequenceNumber()
// 22-Apr-99  FDM added user doc (commented tags)
// 22-Apr-99  M.C added getDefaultData() to create context data from System properties
// 27-Apr-99  M.C added convenience functions (setCycleType(),...,setTimeOut())
//
//-----------------------------------------------------------------------------

package cdev;

import cdev.data.*;

import java.util.*;

/**
 * The DeviceContext class provides the encapsulation into an object of the parameters 
 * that condition the execution of the I/O methods (analogous to a graphics context for screen I/O).
 * Devices can (and do by default) share the same DeviceContext.
 *
 * A default DeviceContext is provided for an "application", based on the thread group name.
 *
 * The DeviceContext is implemented as a cdev.Data, a collection of name-value pairs, and a context
 * number which changes whenever any of the entries change.
 *
 * This class includes a static device context which has a number of flags (entries with value=1)
 * preset to force a read of a property to also retrieve a large number of characteristics 
 * (limits, units, etc.).
 *
 * Convenience routines have been created to manipulate a number of commonly used flags:
 * <P>
 * <TABLE BORDER WIDTH="100%" NOSAVE > 
 * <TR NOSAVE> <TD NOSAVE><B>Valid tags</B></TD> <TD><B>Data type</B></TD> <TD><B>Usage</B></TD> </TR>
 * <TR><TD>cycleType</TD><TD>String or int</TD><TD>Cycle type condition.</TD></TR>
 * <TR><TD>timingEvent</TD><TD>String</TD><TD>Timing event.</TD></TR>
 * <TR><TD>timeOut</TD><TD>double</TD> <TD>A time out value for I/O methods execution, in seconds.</TD></TR>
 * <TR><TD>count</TD><TD>int</TD><TD>User-specified number of values to return.</TD></TR>
 * <TR><TD>referenceKey</TD><TD>String</TD><TD>Reference value set identifier</TD></TR>
 * <TR><TD>onChange</TD><TD><I>any</I></TD><TD>If present, the "on-change" mode is active for monitors</TD></TR>
 * </TABLE>
 * <P>
 * Constructors use the following system properties: (this list should later be made configurable)
 * <UL>
 * <LI> cdev.context.cycleType
 * <LI> cdev.context.onChange
 * </UL>
 */
public class DeviceContext {

  //-------------------------------------------------------------//
  // DeviceContext Table:
  //-------------------------------------------------------------//
  private static final int   HASHTABLESIZE = 20;
  private static final float HASHLOAD = (float) 0.25;
  private static Hashtable deviceContextTable = new Hashtable(HASHTABLESIZE,HASHLOAD);

  private static int nextVersion = 1;

  Data data;
  int contextVersion;

  public DeviceContext () {
    data = getDefaultData();
    contextVersion = nextVersion++;
  }

  public DeviceContext (Data d) {
    if (d == null) 
      data = getDefaultData();
    else
      data = new Data (d); // !!! this must be replaced by a deep copy!!!
    contextVersion = nextVersion++;
  }

  public static void setDefault (DeviceContext ctx) {
    String name = getDefaultName ();
    setDefault (name,ctx);
  }

  public static void setDefault (String name, DeviceContext ctx) {
    if (ctx == null) ctx = new DeviceContext();
    if (name == null) name = getDefaultName ();
    deviceContextTable.remove(name);
    deviceContextTable.put(name,ctx);
  }

  public static DeviceContext getDefault () {
    String name = getDefaultName ();
    return getDefault(name);
  }
    
  public static DeviceContext getDefault (String name) {
    if (name == null) name = getDefaultName ();
    DeviceContext ctx = (DeviceContext) deviceContextTable.get(name);
    if (ctx == null) {
      ctx = new DeviceContext();
      deviceContextTable.put(name,ctx);
    }
    return ctx;
  }
    
  // helper routine to get the default control system name,
  // currently the name of the current thread group
  static String getDefaultName() 
  {
    return ControlSystem.getDefaultName();
  }

  public void setData (Data d) {
    contextVersion = nextVersion++;
    data = new Data(d); // should be a clone (deep copy)
  }

  public Data getData () {
    return new Data(data); // again, should be a deep copy
  }

  // method available to IOService for optimization of access/update
  public int getSequenceNumber() {return contextVersion;}

  //method to get a default Data object from System properties
  //!!! thie list of flags being set from system properties should be dynamic!!!
  private Data getDefaultData() {
    Data d = new Data();
    try {
      Properties props = System.getProperties();
      String cycleType = props.getProperty("cdev.context.cycleType");
      if (cycleType != null)
	d.insert("cycleType",cycleType);
      String onChange = props.getProperty("cdev.context.onChange");
      if (onChange != null)
	d.insert("onChange", onChange);
    } catch (Exception e) {}
    return d;
  }

  // Convenience functions

  /**
   * Stores a cycle type selection into this DeviceContext.
   * Checks first if new selection.
   */
  public void setCycleType (String cycleType) {
    int ct_tag = Data.convertKey ("cycleType");
    setStringSelection(ct_tag,cycleType);
  }
  /**
   * Stores a timing event selection into this DeviceContext.
   * Checks first if new selection.
   */
  public void setTimingEvent (String timingEvent) {
    int te_tag = Data.convertKey ("timingEvent");
    setStringSelection(te_tag, timingEvent);
  }
  /**
   * Stores a referenceKey (archive name) selection into this DeviceContext.
   * Checks first if new selection.
   */
  public void setReferenceKey (String referenceKey) {
    int rk_tag = Data.convertKey ("referenceKey");
    setStringSelection(rk_tag, referenceKey);
  }
   /**
   * Stores a timeOut value into this DeviceContext.
   * Checks first if new value.
   */
  public void setTimeOut (double timeOut) {
    int to_tag = Data.convertKey ("timeOut");
    DataEntry de = data.getDataEntry (to_tag);
    // check if new value
    if (de != null) {
      if (de.doubleValue() == timeOut)
	return;
    }
    // store new value
    data.addDataEntry(new DataEntry(to_tag, timeOut));
    contextVersion = nextVersion++;
  }

  private void setStringSelection(int dataEntryTag, String value) {
    
    DataEntry de = data.getDataEntry (dataEntryTag);
    // check if new value
    if (de != null) {
      String val = de.stringValue();
      if (val.equals(value))
	return;
    }
    // store new value
    data.addDataEntry(new DataEntry(dataEntryTag, value));
    contextVersion = nextVersion++;

  }
}




