//-----------------------------------------------------------------------------
//
//  Description:
//      Transaction class for cdev
//
//  Author: Chip Watson
//
//-----------------------------------------------------------------------------
// Revision History:
//
// 11-Aug-98  CW Prototype version
//  3-Sep-98  FDM opcodes & getTransactionNumber() made public.
//  5-Nov-98  CW added outbound field and changed constructor
//  9-Nov-98  FDM added isSynchronous()
// 23-Nov-98  FDM added GETREFERENCE, SETREFERENCE opcodes, extended codeNames
//  3-Dec-98  CW added accessor for outbound field
//  8-Apr-99  M.C changed class from package cdev to cdev.io, modified access to context
//  3-Jun-99  CW allow inheritance for ClipTransaction
//-----------------------------------------------------------------------------

package cdev;

import cdev.*;
import cdev.data.*;
import java.util.*;

public class DeviceTransaction extends DeviceEvent {

  // opcodes 
  public static final int GET = 1;
  public static final int SET = 2;
  public static final int MONITORON = 3;
  public static final int MONITOROFF = 4;
  public static final int SEND = 5;
  public static final int GETREFERENCE = 6;
  public static final int SETREFERENCE = 7;

  public static final int MAXCODE = 7; // should not be public ?

  protected static final String[] codeNames = 
  { "invalid",
    "get",
    "set",
    "monitorOn",
    "monitorOff",
    "send",
    "getReference",
    "setReference"
  };

  // access levels to review after packages re-org.
  // need public for ClipService
  public static final int PENDING       = 1;
  public static final int INCOMPLETE    = 2;
  public static final int RECEIVED      = 3; // service done, event not delivered
  public static final int COMPLETE      = 4;
  public static final int CANCELLED     = 5;

  static int transCnt = 1;

  protected Data outbound;
  protected Object listener;
  protected int transactionNumber;
  protected int contextVersion;

  protected DeviceTransaction (Device d, int opcode, String arg, Data out, DeviceListener lis) {
    super (d,opcode,arg,null,null);
    outbound = out;
    listener = lis;
    contextVersion = d.getContext().getSequenceNumber();
  }

  DeviceTransaction (Device d, int opcode, String arg, Data out, Data results) {
    super (d,opcode,arg,results,null);
    outbound = out;
    contextVersion = d.getContext().getSequenceNumber();
  }

  // null constructor s.t. derived classes may override
  // used when transaction will not hold a device instance
  public DeviceTransaction() {
    super (null,0,null,null,null);
  }

  public int getState() {return state;}

  public int getTransactionNumber() {return transactionNumber;}

  public Data getOutboundData() { return outbound;}
  public Data getRequestData() { return outbound;}

  public void setData(Data d) {data = d;}

  public void setError(DeviceError err) { error = err; }

  void setIndex(int index) { transactionNumber = makeTxNum(index); }

  int getIndex() { return extractIndex(transactionNumber); }

//---------------------------------------------------------------------//
// makeTxNum:
// Helper function to generate transaction numbers.  Transaction numbers
// are made by concatenating a transaction count and an index into
// the table where the transaction is entered.  The transaction count
// "transCnt" has to be in the high 2 bytes and transaction number must be
// greater than zero.
//---------------------------------------------------------------------//
  private static final int makeTxNum (int index)
  {
    int txnum;
    transCnt++;         // transCnt is never ever decremented.
    txnum = (transCnt << 16) + index;
    if (txnum < 0) {
      transCnt = 1;	// assure positive transaction number
      txnum = (transCnt << 16) + index;
    }
    return (txnum);
  }


  //**********************************************************************************//
  /** extractIndex:
   * Extract the index from the transaction number.  
   *
   */
  static final int extractIndex (int transNum)
  {
    return (transNum & 0x0000FFFF);
  }

  //**********************************************************************************//
  /**
   * addListener:
   * Transactions are combined if they request the same service.  This 
   * saves network bandwidth and server processing time. 
   */
  final void addListener(Object obj)
  {
    if (listener == null) {
      listener = obj;
      return;
    }
    if ( !( listener instanceof Vector ) ) {
      Vector txVec = new Vector(); // convert to a vector of listeners
      txVec.addElement(listener);
      listener = txVec;
    }
    ((Vector)listener).addElement(obj); // add the new listener
  }

  //**********************************************************************************//
  /**
   * removeListener:
   * This removes the DeviceListener or a Data from a vector.
   */
  public void removeListener(Object obj)
  {
    if (listener == obj) {
      listener = null; // removed last one
      return;
    }
    if (listener==null || !(listener instanceof Vector) ) return; // safety

    Vector list = (Vector) listener;
    while (list.removeElement(obj)) {}

    if (list.size() == 0) listener = null;
  }

  //**********************************************************************************//
  /**
   * isSynchronous:
   * synchronous means Data and no listener.
   */

   public boolean  isSynchronous()
   {
     return (listener == null);
   }
}
