//---------------------------------------------------------------------------
// Copyright (c) 1996,1997 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description:
//      This class manages transactions in a table, and handles monitoring
//      functions -- collapsing duplicates and firing monitors on request,
//      as responses arrive from the IOservice.  
//
//      A single transaction may hold multiple listeners, as when multiple
//      GUI objects display the same value.
//
//      Multiple transactions may be held at the same slot in a linked list,
//      as when multiple clients initiate equivalent monitor requests.
//
//  Author: Chip Watson
//
//  Partially based on work by Walt Akers, Jie Chen, and Cuong Quach plus ideas 
//  from Franck DiMaio at CERN
//                                       
//----------------------------------------------------------------------------
// Revision History:
//
// 23-Apr-99 CW Created from ControlSystem.java
// 28-May-99 CW Added linked list to support future monitoring within servers
//
//----------------------------------------------------------------------------

package cdev;

import cdev.data.*;
import cdev.io.*;

import java.util.*;

/**
 *
 * @version 2.1
 * <br>
 * Class to manage CDEV transactions.
*/
public class TransactionManager
{
  //-------------------------------------------------------------//
  // txTable:
  // Private data for implementing a table of transactions.
  //-------------------------------------------------------------//
  private static final int    INIT_TABLE_SIZE = 100;
  private static final float  LOAD_FACTOR = (float)0.9; // % filled before growing

  private DeviceTransaction[] txTable = new DeviceTransaction[INIT_TABLE_SIZE];
  private Object tableLock = new Object(); // something to lock (not array)
  private int free = 0;
  private int occupancy = 0;
  private int margin = (int) (LOAD_FACTOR * INIT_TABLE_SIZE);
  private int activeCnt = 0;        // count of active transactions for this system


  //**********************************************************************************//
  /**
   * Constructor.
   */
  public TransactionManager ()
  {
      // data initialized above
  }


  //**********************************************************************************//
  /**
    * add:
    * Add a transaction into the transaction table
    */
  public final void add ( DeviceTransaction dt ) {
    int index = getFree();
    dt.setIndex(index);
    txTable[index] = dt;
    activeCnt++;
    dt.state = DeviceTransaction.PENDING;
  }

  //**********************************************************************************//
  /**
    * elements:
    * Return an enumeration of transactions
    */
  public final Enumeration elements ( ) {
      class TmEnumeration implements Enumeration {
	int next;
	int tlen;
	TmEnumeration() {
	  next = 0;
	  tlen = txTable.length - 1;
	}
	public boolean hasMoreElements() {
	  while (next < tlen ) {
	    if (txTable[next]!=null) break;
	    next++;
	  }
	  return (txTable[next]!=null);
	}
	public Object nextElement() {
	  if (txTable[next]!=null)
	    return txTable[next++];
	  if (!hasMoreElements()) throw new NoSuchElementException();
	  return txTable[next++];
	}
      }
    return new TmEnumeration();
  }

  //**********************************************************************************//
  /**
    * addListener:
    * Add a listener to an existing transaction. If no matching monitor found,
    * return false.
    *
    * The last argument is an Object so that this method can handle sendNoBlock calls.
    * The transactionHandler will use the instanceof operator to determine what
    * this is and carry out the appropriate action.  If it is an instanceof Data
    * the handler only has to copy the result over.  If it is an instanceof 
    * DeviceListener, then the transactionHandler will call the appropriate method
    * of that object.
    */
  final boolean addListener (String devname, String property, 
			     int cv, Object listener) {

    DeviceTransaction dt = null;

    boolean result = false;

    // ---------------------------------------------------------------------  
    //  Really want to sync on the transaction table but can't sync on
    //  arrays.  So lock a special object.
    // ---------------------------------------------------------------------  

    if (listener!=null) { 
      synchronized (tableLock) {

        // attempt to merge duplicate monitorOn calls
        dt = findTrans (devname, DeviceTransaction.MONITORON, property, cv);
        if (dt != null)  {                   // transaction found in table
	  dt.addListener (listener);
	  if (dt.data != null) { // reply immediately if cache has data
	    dispatch (dt,listener);
	  }
	  result = true;
	}
      }
    }
    return result;
  }

  //**********************************************************************************//
  /**
    * removeListener:
    * Look for a matching "monitorOn" and if one is found, remove
    * the transaction from txTable if only one pending listener. 
    * Otherwise just remove this listener.  If no
    * matching transactions are found, ignore this call.
    */
  final DeviceTransaction removeListener (String devname, String property, 
					  int cv, Object listener) {
    DeviceTransaction dt;
    dt = findTrans (devname, DeviceTransaction.MONITORON, property, cv);
    if (dt != null)  {                           // transaction found in table
      dt.removeListener(listener);
      if (dt.listener == null) {
	if (dt.state == DeviceTransaction.COMPLETE) activeCnt++;
	removeTrans(dt); // this will decrement activeCnt
	return dt; // return any removed transaction
      }
      dt = null; // this transaction is still in use by another client, don't kill
    }
    return dt;
  }

  //**********************************************************************************//
  /**
   * removeTrans:
   * Remove a transaction from the transaction table.
   */
  final void removeTrans (DeviceTransaction dt)  
  {
    synchronized (tableLock) {
      int index = dt.getIndex();
      if (   (index < txTable.length ) && (txTable[index] == dt )  ) {
	txTable[index] = null;
	occupancy--;
      }
    }
    activeCnt--;
  }

  /**
   * Suspend IO
   */
  final void suspendIO(ControlSystem sys) {

    for (int i=0;i<txTable.length;i++) {
      DeviceTransaction dt = txTable[i];
      if ( dt != null ) try {
	Device dev = dt.getDevice();
	if (dev.getControlSystem() == sys) {
	  IOService service = dev.getIOService();
	  service.cancel(dt);
	}
      } catch (Exception e) {}
    }
  }

  /**
   * Resume IO
   */
  public void resumeIO (ControlSystem sys) {
    for (int i=0;i<txTable.length;i++) {
      DeviceTransaction dt = txTable[i];
      if ( dt != null ) try {
	dt.state = DeviceTransaction.PENDING;
	Device dev = dt.getDevice();
	if (dev.getControlSystem() == sys) {
	  IOService service = dev.getIOService();
	  service.post(dt);
	}
      } catch (Exception e) {
	  System.out.println("resume of "+dt+" failed! "+e);
	  e.printStackTrace();
      }
    }
  }

  //---------------------------------------------------------------------//
  private DeviceTransaction findTrans (String devname, int opcode, String arg, 
				     int cversion)
  {
    synchronized (tableLock) {
      DeviceTransaction dt;
      for (int i=0;i<txTable.length;i++) {
	dt = txTable[i];
	if ( dt != null
	     && dt.opcode == opcode
	     && dt.getDeviceName().equals(devname)
	     && dt.arg.equals(arg)
	     && dt.contextVersion == cversion
	     )
	  return dt;
      }
    }
    return null;
  }

  //**********************************************************************************//
  /**
   *  getFree:
   *  Find a free spot in the table to add an object.  Grow the table
   *  if the occupancy margin is reached and copy the old table contents
   *  to the new table.
   */
  private int getFree ()
  {
    synchronized (tableLock) {

      //-----------------------------------------------------
      // Find next free spot.  Free should never sit at last
      // element in array.  This search will always find an
      // empty slot.
      //-----------------------------------------------------
      while (txTable[free] != null)  free++;

      //-----------------------------------------------------
      // Now the next free spot is found, return it if its not
      // the last entry in the array!
      //-----------------------------------------------------
      if (free < (txTable.length - 1) )  {
	occupancy++;
	return free;
      }
      //-----------------------------------------------------
      // Now handle the wrap case!  Reset free to the beginning 
      // of array and re-search the array if occupancy < margin.
      //-----------------------------------------------------
      if (occupancy < margin)    {   // still plenty room left
	free = 0;
	while (txTable[free] != null)    free++;
	//Note: Bottom check skipped here!  In theory, there
	//      is space in array above original starting point.
	occupancy++;
	return free;
      }

      //System.out.println("free="+free+" occ="+occupancy+
      //		 " margin="+margin);

      //-----------------------------------------------------
      // If code gets here, the txTable has to expand and free
      // should be at the last element in the array.
      //-----------------------------------------------------
      DeviceTransaction[] temp = txTable;
      //--- double the table size & create new table ---
      int capacity = txTable.length * 2;
      margin = (int) (LOAD_FACTOR * capacity);
      txTable = new DeviceTransaction[capacity];
      //System.out.println ("TransactionManager table capacity now "+capacity);

      //--- copy the old stuff over --------------------
      System.arraycopy(temp,0,txTable,0,temp.length);
    }
    occupancy++;
    return free;
  }  

  //*****************************************************************************
  /**
   *  findTransaction:
   *  Called to find a pending transaction based upon the
   *  transaction number (used for bookkeeping).
   */
  public final DeviceTransaction findTransaction(int trnum) {
    int index = DeviceTransaction.extractIndex(trnum);
    if (index >= 0 && index < txTable.length) {
      DeviceTransaction dt = txTable[index];
      if ( (dt != null) && (dt.getTransactionNumber() == trnum) ) {
	return dt;
      }
    }
    return null;
  }

  //---------------------------------------------------------------------
  // process:
  // Dispatch events to listeners.  If just a result data is required, copy
  // the data and wake any waiting threads.  Delete the transaction when 
  // done handling the response.
  //---------------------------------------------------------------------
  public void process (Fifo fifo) {
    DeviceTransaction dt;
    while ( fifo.notEmpty() ) {
      dt = (DeviceTransaction) fifo.next();
      process(dt);
    }
  }
  public void process (DeviceTransaction dt) {
    Object     listener = null;
    Vector     dlList = null;

    if (dt.state == DeviceTransaction.CANCELLED) return; // killed already

    listener = dt.listener;                   // get any listeners

    if (  (listener != null) ) {

      //-------------------------------------------------------
      // Handle transactions with vectored callbacks (only monitors)
      //-------------------------------------------------------
      if (listener instanceof Vector)  { // only possible for monitors
	dlList = (Vector) listener;
	int len = dlList.size();
	//System.out.println ("ControlSystem::run: handle "+len+" listeners...");

	for (int i=0; i<len; i++)  {
	  dispatch (dt,dlList.elementAt(i)); // *** possible race condition ***
	}

	// update the active count for pend()
	if (dt.state != DeviceTransaction.COMPLETE) { // i.e. first data
	  dt.state = DeviceTransaction.COMPLETE; //=> 1st data received
	  activeCnt--; // decrement on first data
	}

      } else {
	//-------------------------------------------------------
	// Handle transactions with a single callback. 
	//-------------------------------------------------------
	dispatch (dt, listener);

	//------------------------------------------------------------------
	// Delete the transaction entry if the associated message is not a 
	// monitorOn.
	//------------------------------------------------------------------

	if (dt.opcode != DeviceTransaction.MONITORON )    {
	  if (dt.state != DeviceTransaction.INCOMPLETE) {
	    //System.out.println("removing "+dt);
	    removeTrans(dt);
	  }
	} else {
	  if (dt.state != DeviceTransaction.COMPLETE) { // i.e. first data
	    dt.state = DeviceTransaction.COMPLETE; //=> 1st data received
	    activeCnt--; // decrement on first data
	  }

	}  // finish cleaning 

      } // single listener

    } // non-null listener
    else { // notify user thread if synchronous transaction FDM
      synchronized (dt) {
	dt.state = DeviceTransaction.COMPLETE; // for late pend(dt)
	dt.notify(); // synchronous case, user is pending on transaction
	removeTrans(dt);
      }
    }
  }  // end of process method


  private void dispatch (DeviceTransaction dt, Object obj) {

    // if no listener, just notify anyone waiting on this transaction
    if (obj == null) {
      synchronized (dt) { dt.notify(); }
    }

    //--------------------------------------------------
    // Send event to listener
    //--------------------------------------------------
    else if (obj instanceof DeviceListener) {
      DeviceListener dl = (DeviceListener) obj;
      //System.out.println ("TransactionManager::dispatch: call listener...");
      dl.deviceChanged (dt); // listener cannot change the data !!!
    } else if (obj instanceof TransactionResponseListener) { // deprecated
      TransactionResponseListener tl = (TransactionResponseListener) obj;
      tl.handleResponse(dt);
    }
  }

  public final int getActiveCnt() {return activeCnt;}

  // *************************************************************************
  //-------------------------------------------------------------
  // Debug methods:
  //-------------------------------------------------------------

  /**
   * Converts a transaction to a string.  The index is the actual index into
   * the transaction table, not the number associated with the specific
   * transaction.  The title is a user given title to be attached to the
   * beginning of the transaction string.
   */
  //private String txToString (String title, int index)
  public String txToString (String title, int index)
  {
    StringBuffer result = new StringBuffer();
    if (title != null)  result.append(title).append("\n");
    DeviceTransaction dt = txTable[index];
    result.append(dt.toString()).append(".");
    result.append(Integer.toHexString(dt.getTransactionNumber()));
    return result.toString();
  }
  
  /**
   * Steps through the transaction table and convert each transaction to 
   * a string.  The Title is attached to the beginning of the string.
   */
  //private String txTableToString (String title)
  public String txTableToString (String title)
  {
    StringBuffer result = new StringBuffer();
    if (title != null)  result.append(title).append(" contents:");

    int     entryhits = 0;
    int     i = 0;
    
    //System.out.println ("there are ["+occupancy+"] transactions in the table.");
    while (entryhits < occupancy) {
      if ( txTable[i] != null )  {
        result.append(txToString ("\ntx["+i+"]",i));  
        entryhits++;
      }
      i++; 
      //System.out.println ("print check txTab["+i+"]");
    }
    return result.toString();
  }

}


