//-----------------------------------------------------------------------------
//
//  Description:
//      CLIP I/O Service.  This is the interface between
//      the re-usable parts of the cdev implementation and
//      the CLIP protocol.  It supports both client and
//      server connections, with different constructors
//      for each.  Outbound messages are buffered, and sent
//      when explicitly flushed, or after a timeout.
//
//  Author: Chip Watson
//  (originally based upon work by Cuong Quach)
//
//  This class extends Thread to continuously read the socket.
//
//-----------------------------------------------------------------------------
// Revision History:
//
//   9-Oct-98  CW Prototype version
//  29-Apr-99  CW Modified to support clients and servers
//  11-Jan-00  CW Added context transmission on change
//   3-Feb-00  CW Fix bug in cancel of monitor
//-----------------------------------------------------------------------------

package cdev.clip;

import cdev.*;
import cdev.data.*;
import cdev.io.*;
import cdev.dir.*;

import java.io.*;
import java.net.*;
import java.util.*;

public class ClipService 
    implements Runnable, IOService, TransactionResponseListener {

  // singleton instance for callers of getIOService()
  private static ClipService singleton = null;
  private static int nextClientID = 1;

  private static final int MAGIC_NUMBER = 0x434c4950; // part of buffer header, "clip"
  private static final int MAXERROR = 10;             // number of words to scan for magic#
  private static final int MAXPACKET = 100;           // maximum packets per CLIP buffer

  private static final int DEFAULT_OUTPUT_BUFFER_SIZE = 16000;
  private static final int WRITE_FLUSH_THRESHOLD = 8000; // above this, auto flush
  private static final int FLUSH_TIMEOUT = 100; // milliseconds to auto flush

  // the packet revision
  // private static final char VERSION = '\u0001'; // before C++ cdev 1.6.2
  private static final char VERSION = '\u0002';    // for C++ cdev 1.6.2 

  // Clip Header bitmap
  private static final char CLIENT_ID_SET          = '\u8000';
  private static final char TRANS_INDEX_SET        = '\u4000';
  private static final char CANCEL_TRANS_INDEX_SET = '\u2000';
  private static final char LOCAL_DATA_INDEX_SET   = '\u1000';
  private static final char FOREIGN_DATA_INDEX_SET = '\u0800';
  private static final char OPERATION_CODE_SET     = '\u0400';
  private static final char COMPLETION_CODE_SET    = '\u0200';  // C++ cdev 1.6.2 format
  private static final char DEVICE_LIST_SET        = '\u0100';
  private static final char MESSAGE_SET            = '\u0080';
  private static final char DATA_SET               = '\u0040';
  private static final char CONTEXT_SET            = '\u0020';
  private static final char TAG_MAP_SET            = '\u0010';

  // connection information
  private String hostname;
  private int portnum;
  private Socket hostSock;
  private boolean isserver;
  private TransactionListener listener;
  private boolean isstarting;

  // input
  private BufferedInputStream inStream;
  private xdrReader reader;
  private Thread readerThread;

  // output
  private OutputStream outStream;
  private DataOutputStream outWriter;
  private ByteArrayOutputStream streamBuffer;
  private DataOutputStream bufferWriter;
  private ByteArrayOutputStream transBuffer;
  private xdrWriter transWriter;
  private int packetCnt;
  private int clientID;
  private String [] names;
  private int lastContextNumber;

  // auto flush
  private Thread auto;

  //------------------------------------------------------------------------//

  /** 
   * Default constructor is used by clients to create a
   * connection to a server using a subsequent connect call.
   *
   * Incoming traffic is passed to cdev.ControlSystem,
   * assuming that it is a response to a Device I/O call.
   */
  public ClipService () {
    isserver = false;
    isstarting = false;
  }

  /**
   * Construct a CLIP I/O service connected to the specified
   * socket, and with incoming traffic directed to the specified
   * listener. This form is used by servers creating new
   * client handlers.
   */
  public ClipService (Socket s, TransactionListener l) throws IOException {
    isserver = true;
    isstarting = false;
    hostSock = s;
    hostname = s.getInetAddress().getHostName();
    portnum = s.getPort();
    listener = l;
    listener.registerIOService(this);
    initialize();
  }

  /**
   * Return an instance of this I/O service. In this
   * case, a static instance is returned for use by
   * the Device class.
   */
  public static IOService getIOService() {
    if (singleton == null) singleton = new ClipService();
    return singleton;
  }

  /**
   *  post:
   *  Post a transaction. First, the transaction is serialized into
   *  an output buffer. If a byte threshold is then exceeded, the
   *  buffer is transmitted. The buffer is also transmitted if a
   *  maximum number of transactions have been buffered.
   *
   *  Behavior is a function of transaction state, used to distinguish
   *  between new client requests, and server responses
   */
  public synchronized void post (DeviceTransaction dt) throws IOException {

    // (initially derived from old Packet.streamOut)
    String tempMsg = null;
    int map;
    boolean failed = false;
    DeviceError deverr = null;
    Data data;
    Data context = null;

    testConnect();

    // map indicates which optional fields are present (being sent)
    // map is a function of whether this is a new request being transmitted,
    // or is a response to a request
    //
    // future improvements possible:
    //   1) send client_id only on first request message
    //   2) use device_list_id's after first transmission of device name(s)

    int state = dt.getState();
    
    if (state == DeviceTransaction.PENDING) { // new transaction

      map = (VERSION << 16) |CLIENT_ID_SET |TRANS_INDEX_SET |DEVICE_LIST_SET;

      int opcode = dt.getOpcode();

      if (opcode == DeviceTransaction.SEND) {
	tempMsg = dt.getMessage();
	map |= MESSAGE_SET;

      } else {

	// map new style get/set to old style send
	// someday extend protocol to carry these directly
	if (opcode == DeviceTransaction.GET) {
	    tempMsg = "get " + dt.getProperty();
	    opcode = DeviceTransaction.SEND;
	    map |= MESSAGE_SET;
	} 
	else if (opcode == DeviceTransaction.SET) {
	    tempMsg = "set " + dt.getProperty();
	    opcode = DeviceTransaction.SEND;
	    map |= MESSAGE_SET;
	}
	else if (opcode == DeviceTransaction.MONITORON) {
	    tempMsg = "monitorOn " + dt.getProperty();
	    opcode = DeviceTransaction.SEND;
	    map |= MESSAGE_SET;
	}
	else if (opcode == DeviceTransaction.MONITOROFF) {
	    tempMsg = "monitorOff " + dt.getProperty();
	    opcode = DeviceTransaction.SEND;
	    map |= MESSAGE_SET;
	}
      }
      data = dt.getOutboundData();

      // optionally send context for this I/O operation if context has
      // changed since last transmission
      DeviceContext ctx = dt.getDevice().getContext();
      if (ctx.getSequenceNumber() != lastContextNumber) {
	context = ctx.getData();
	map |= CONTEXT_SET;
      }


    } else { // state not PENDING, so must be a response
      map = (VERSION << 16) |CLIENT_ID_SET |TRANS_INDEX_SET;
      data = dt.getData();
      deverr = dt.getError();
      if (deverr != null) {
	failed = true;
	data = toData(deverr);
	map |= COMPLETION_CODE_SET;
      }
    }

    if (data != null) map |= DATA_SET;

    transWriter.put(map);

    if (state == DeviceTransaction.PENDING) { // client request

      transWriter.put(clientID);
      transWriter.put(dt.getTransactionNumber());

      // always send array of names (even if only one name)
      Device dev = dt.getDevice();
      if (dev instanceof DeviceArray) {
	String [] temp = null;
	try {
	  temp = ((DeviceArray)dev).getNames();
	} catch (DataNotFoundException e) {
	  temp = names;
	  names[0] = dev.getName();
	}
	transWriter.put(temp);
      } else {
	names[0] = dev.getName();
        transWriter.put(names);
      }
      transWriter.put(tempMsg);

    } else {             // send status of response
      ClipTransaction ct = (ClipTransaction) dt;
      transWriter.put(ct.clientID);
      transWriter.put(ct.clienttrans);
      if (failed) transWriter.put(ct.getError().getCode());
    }

    if (data != null) transWriter.put(data.xdrExport());

    if (context != null) transWriter.put(context.xdrExport());

    // transfer serialized transaction to output buffer
    // pre-pending the byte length
    int transLen = transBuffer.size();
    bufferWriter.writeInt (transLen);
    transBuffer.writeTo (streamBuffer);
    transBuffer.reset();

    packetCnt++;

    if (streamBuffer.size() > WRITE_FLUSH_THRESHOLD 
	|| packetCnt >= MAXPACKET) {
      doFlush();
    }
  }

  /**
   * Handle a response to a transaction originated by
   * an incoming request from this object. (Server).
   * This interface is defined by the interface
   * TransactionResponseListener.
   */
  public void handleResponse(DeviceTransaction ev) {
    ClipTransaction ct = (ClipTransaction) ev;
    if (ct.getState() == DeviceTransaction.PENDING) {
      ct.setState(DeviceTransaction.COMPLETE);
    }
    try {
      post (ct);
    } catch (Exception e) {
      cleanup(); // this connection now useless
    }
  }

  /**
   * Flush any buffered I/O.
   */
  public synchronized void flush () throws IOException {
    if (packetCnt > 0) {
      testConnect();
      doFlush();
    }
  }

  private void doFlush() throws IOException {
    int dataLen = streamBuffer.size() + 4;	// length includes packet count
    //System.out.println("flushing "+dataLen+" bytes, "+packetCnt+" packets");
    try {
      outWriter.writeInt(MAGIC_NUMBER);		// first, the magic number
      outWriter.writeInt(dataLen);		// buffer length
      outWriter.writeInt(packetCnt);		// packet count
      streamBuffer.writeTo(outWriter);		// send byte buffer to socket
      streamBuffer.reset();			// reset "pointer" to beginning of buffer
      outWriter.flush();
      packetCnt = 0;
    } catch (SocketException e) {
      cleanup();				// give up for now
    }
  }

  /**
   * Cancel a transaction. Currently converts monitorOn to
   * monitorOff. Others are ignored.
   */
  public void cancel (DeviceTransaction dt) throws IOException {

    if (!isConnected()) return;			// quietly ignore late cancels
    if (dt.getOpcode() == DeviceTransaction.MONITORON) {
	ClipTransaction temp;
	temp = new ClipTransaction(dt.getDevice(),
				   DeviceTransaction.MONITOROFF,
				   dt.getProperty() );
	temp.setState(DeviceTransaction.PENDING);
	post(temp);
    }
  }

  /** 
   * The run method continuously reads the socket for 
   * inbound messages. Messages are forwarded to either 
   * cdev.ControlSystem (client mode) or to the listener 
   * specified in the constructor.
   */
  public void run() {
    int     thisLoad        = 0;
    int     overAllLen      = 0;
    int     magicNum        = 0;
    int     errorCnt        = 0;

    isstarting = false;
    //System.out.println("CLIP reader starting...");
    while (true) {
      try {
  
        //----------------------------------------------------------//
      	// Reset count before every message/frame from server 
        //----------------------------------------------------------//
      	magicNum = 0;
	thisLoad = 0;             
      	overAllLen = 0;
      
        //----------------------------------------------------------//
      	// Block here waiting for next message/frame from server.  
        //----------------------------------------------------------//
	// Read magic number & frame length from the socket.
        //----------------------------------------------------------//
	magicNum = reader.get_int();
	//System.out.println("CLIP:: magic ["+magicNum+"] errorCnt "+errorCnt);

      	if (magicNum != MAGIC_NUMBER)  {
	  errorCnt++;
	  if (errorCnt <= MAXERROR) {
	    if (errorCnt == 1) 
	      System.out.println ("CLIP::bad magic number ["+
			        Integer.toHexString(magicNum)+"] scanning...");
	    else
	      System.out.print ("... "+Integer.toHexString(magicNum));
	    continue;
	  } else {
	    disconnect();
	    break; // exit while loop
	  }
      	}
	if (errorCnt > 0) {
	  System.out.println("recover after errorCnt = "+errorCnt);
	  errorCnt = 0;	// got the magic number
	}

        //----------------------------------------------------------//
	// Extract frame byte count & check validity 
        //----------------------------------------------------------//
      	overAllLen = reader.get_int(); 
      	//System.out.println ("socketReader:: frame length "+overAllLen);

      	if (overAllLen < 1) {
	  System.out.println("CLIP invalid buffer length: "+overAllLen);
	  continue;     // start searching for magic number again
      	}

        //----------------------------------------------------------//
	// Read the remaining data for the frame.
        //----------------------------------------------------------//
	// Extract packet count & check validity.
        //----------------------------------------------------------//
	thisLoad = reader.get_int();
        //System.out.println("socketReader:: #packets "+thisLoad);
	if (thisLoad < 1) {
	  System.out.println("socketReader::Invalid packet count: "+thisLoad);
	  continue;
	}

        //----------------------------------------------------------//
	// Extract each packet and insert into fifo queue.
        //----------------------------------------------------------//
	for (int i=0; i<thisLoad; i++) {
	  handlePacket();
	}

      } catch (Exception e) {
	System.out.println("CLIP exception "+e.getClass().getName());
	// e.printStackTrace();
        //----------------------------------------------------------//
	// Error handling for broken socket exceptions.
        //----------------------------------------------------------//
        //if (e.getClass().getName().equals("java.io.EOFException")) {
	if (isserver)
	  cleanup(); // on any error, discard this connection
	else
	  disconnect(); // for now just disconnect client connections
	//}
      }  // end of exception handling

    }  // repeat forever
  }

  /** 
   * Handle a single packet from within a block of packets.
   */
  private void handlePacket() throws IOException {
    // from old Packet.streamIn()
    int id = 0;
    int transIndex = 0;
    int cancelTransIndex;
    int localDataIndex;
    int foreignDataIndex;
    int operationCode;
    int completionCode = 0;
    String[] deviceList = null;
    String message = null;
    Data data = null;
    DeviceError error = null;
    Data ctx = null;
    Data tagMap;

    //----- get and discard the length (for now) --------//
    reader.get_int();

    //----- get the version number and packet map -------//
    int packetMap = reader.get_int();
    int version = packetMap >> 16;
    packetMap <<= 16;

    //----- check version number ------------------------//
    if (version != VERSION)   {
       throw new IllegalArgumentException("Packet version missmatch");
    }

    //----- load elements while reading off the flags ---//
    // NOTE: 
    //   1) This will alter packetMap as it is read in.
    //   2) The sequence that data is read from the buffer 
    //      is based on the definition of CLIP.
    //----------------------------------------------------//
    if (packetMap < 0)  id = reader.get_int();
    packetMap <<= 1;
    if (packetMap < 0)  transIndex = reader.get_int();
    packetMap <<= 1;
    if (packetMap < 0)  cancelTransIndex = reader.get_int();
    packetMap <<= 1;
    if (packetMap < 0)  localDataIndex = reader.get_int();
    packetMap <<= 1;
    if (packetMap < 0)  foreignDataIndex = reader.get_int();
    packetMap <<= 1;
    if (packetMap < 0)  operationCode = reader.get_int();
    packetMap <<= 1;
    // next bit is C++ cdev 1.6.2 message format addition
    if (packetMap < 0)  completionCode = reader.get_int();
    packetMap <<= 1;
    if (packetMap < 0) {
       int Count = reader.get_int();
       deviceList = reader.get_string (Count);
    } else {
	deviceList = new String[1];
	deviceList[0] = "unspecified";
    }
    packetMap <<= 1;
    if (packetMap < 0)  {
      message = reader.get_string();
      //System.out.println ("got msg: "+message);
    }
    packetMap <<= 1;
    if (packetMap < 0)  {
      data = new Data(reader);
      //System.out.println("got data:"+data);
    }
    packetMap <<= 1;
    if (packetMap < 0)  {
      ctx = new Data ( reader );
    }
    packetMap <<= 1;
    if (packetMap < 0)  {
      tagMap = new Data ( reader );
    }
    // If this object is dealing with unsolicited requests from
    // remote clients, create a new transaction and send it to
    // this object's listener.
    if (isserver) {
      ClipTransaction ct;
      StringTokenizer st = new StringTokenizer(message);
      String verb = st.nextToken();
      String pname = st.nextToken();
      int opcode = DeviceTransaction.SEND;
      if (verb.equals("get")) { opcode = DeviceTransaction.GET; }
      else if (verb.equals("set")) { opcode = DeviceTransaction.SET; }
      else if (verb.equals("monitorOn")) { opcode = DeviceTransaction.MONITORON; }
      else if (verb.equals("monitorOff")) { opcode = DeviceTransaction.MONITOROFF; }
      if (opcode != DeviceTransaction.SEND)
	message = pname;

      ct = new ClipTransaction(id,transIndex,deviceList,
			       opcode,message,data,ctx,this);
      listener.post(ct); // post new request from client
    } 

    // If this object is serving as a link between local clients
    // and a remote server, send the response back to ControlSystem.
    else {

      //System.out.println("clip handleResponse "+
      //     Integer.toHexString(transIndex)+" "+
      //     completionCode +" "+ data);
      if (completionCode != 0) {
	DeviceTransaction dt = ControlSystem.findTransaction(transIndex);
	error = new DeviceError(dt,completionCode,
				"from clip server "+hostname+":"+portnum);
      }
      ControlSystem.handleResponse(transIndex, completionCode, 
				   data, error); // no errors
    }
  }


  //
  // routines for connection control (client operations)
  //
  public void connect (String host, int port) throws Exception {

    lastContextNumber = -1;
    clientID = nextClientID++;
    isstarting = true;
    try {
      disconnect();	// tolerate multiple connect calls

      // open the socket
      hostname = host;
      portnum = port;
      hostSock = new Socket (hostname,portnum);
      initialize();

    } catch (IOException e) {
      disconnect();
      isstarting = false;
      throw e;
    }
  }

  public void reconnect () throws Exception {
    if (hostname != null && portnum != 0) {
      connect(hostname,portnum);
    }
  }

  public void initialize () throws IOException {
    System.out.println("CLIP connected to "+hostname+" "+portnum+"...");
    try {
      // set up the reader
      inStream = new BufferedInputStream(hostSock.getInputStream());
      reader = new xdrReader(inStream);
      readerThread = new Thread(this);
      try {readerThread.setDaemon(true);} catch (Exception e) {}
      readerThread.start();

      // set up the writer
      outStream = hostSock.getOutputStream();
      outWriter = new DataOutputStream(outStream);
      streamBuffer = new ByteArrayOutputStream(DEFAULT_OUTPUT_BUFFER_SIZE);
      bufferWriter = new DataOutputStream(streamBuffer);
      transBuffer  = new ByteArrayOutputStream(DEFAULT_OUTPUT_BUFFER_SIZE);
      transWriter = new xdrWriter(transBuffer);
      names = new String[1]; // holds the device name (1 only for now)

      // set up auto flush thread (inner class)
      auto = new Thread() {
	public void run() {
	  while (true) {
	    try {
	      Thread.sleep(FLUSH_TIMEOUT);
	      flush();
	    } catch (Exception e) {}
	  }
	}
      };
      try {auto.setDaemon(true);} catch (Exception e) {}
      auto.start();
    } catch (IOException e) {
      disconnect();
      throw e;
    } catch (Exception e) {e.printStackTrace(); throw new IOException(e.toString());}
  }

  public void disconnect () {

    // *** need to add auto send of sign-off message !!!

    // stop the auto flush thread
    if (auto != null) {
      auto.stop();
      try {auto.setDaemon(false);} catch (Exception e) {}
      auto = null;
    }

    if (readerThread != null) {
      readerThread.stop();		// stop the read thread
      try {readerThread.setDaemon(false);} catch (Exception e) {}
      readerThread = null;
    }

    reader = null;
    inStream = null;
    try {
      if (hostSock != null) {	// clean-up
	hostSock.close();
	hostSock = null;
      }
    } catch (Exception e) {}	// ignore problems
  }

  public final boolean isConnected () {
    return (hostSock != null);
  }

  public String getHostName () {return hostname;}
  public int getPort () {return portnum;}

  private final void testConnect() throws IOException {

    if (isstarting) {
	for (int i=1;i<10;i++) {
	    try {Thread.sleep(10);} catch (Exception e){}
	    if (!isstarting) break;
	}
    }
    if (!isConnected()) {
      throw new IOException("CLIP service not connected!");
    }
  }

  private Data toData(DeviceError err) {
    Data result = new Data();
    result.insert("error",err.getMessage());
    return result;
  }

  private void cleanup () {
    System.out.println("Comm error, disconnecting from "+hostname+" "+portnum);
    if (listener != null) listener.unregisterIOService(this);
    disconnect();
  }

  protected void finalize() throws Throwable {
    cleanup();
    super.finalize();
  }

}
 
