//---------------------------------------------------------------------------
// Copyright (c) 1996,1997 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//---------------------------------------------------------------------------
//
//  Description:
//      Java implementation of a data entry container class
//
//  Issues:
//	1) C++ CDEV Bounds offsets not supported, only dimensions
//      3) In xdrImport, packing byte buffer for network transfer can be
//         more efficient if numElements not sent for dataEntries that only
//         contain one data element (if dim = 0, numElements is 
//         implicitly 1)
//
//  To Do:
//	1) Modify / improve read/write of multi dimensional arrays
//	2) evaluate use of array instead of vector (assumes small number
//	   of entries)
//
//  Author: Chip Watson
//
//                                       
//---------------------------------------------------------------------------
// Revision History:
//
//  9-22-96: Add xdrImport, xdrExport, and supporting methods.
// 10-21-96: Add getType method.
//  2-10-97: Remove cdev prefix from class name references.
//  1-15-98: Remove convenience insert/get functions (shrink class)
//  9-03-98: FDM re-copy insert, add copyInto, Data(xdrReader) and xdrExport public.
// 11-06-98: M.C add copyInto
// 14-Jan-99: CW Fixed xdrImport of scalars, xdrExport of scalar String
// 27-Apr-99: M.C added static public int VALUE_TAG.
//  1-Jun-99: CW Fixed xdrExport for String
//
//---------------------------------------------------------------------------

package cdev.data;

import java.io.*;
import java.util.*;

/**
 * A container class for storing <A href="cdev.data.DataEntry.html">DataEntry</A>
 * objects accessed via either String or int keys.
 */

public class Data implements Cloneable{

  // Enumeration of primitive data types supported by this package

  static public final short INVALID= 0;
  static public final short BYTE   = 1;
  static public final short SHORT  = 2;
  static public final short INT    = 3;
  static public final short LONG   = 4;
  static public final short FLOAT  = 5;
  static public final short DOUBLE = 6;
  static public final short STRING = 7;
  static public final int MAXTYPE = STRING;

    static private final String A_STRING = new String();
    static final Class[] localTypeToClass = { null, Byte.TYPE, Short.TYPE,
					      Integer.TYPE, Long.TYPE, Float.TYPE,
					      Double.TYPE, A_STRING.getClass()};

  // Constants for DataEntry key (see TagTable for corresponding value)
  static public final int VALUE_KEY = convertKey("value");

  // Constants for readability in xdrImport, xdrExport, and
  // supporting methods. Unsigned types converted to wider types on read.
  // Must match cdev C++ implementation for network interoperability.

  static final short CDEV_BYTE       = 0;   
  static final short CDEV_INT16      = 1;   
  static final short CDEV_UINT16     = 2;   
  static final short CDEV_INT32      = 3;
  static final short CDEV_UINT32     = 4;
  static final short CDEV_FLOAT      = 5; 
  static final short CDEV_DOUBLE     = 6; 
  static final short CDEV_STRING     = 7; 
  static final short CDEV_TIMESTAMP  = 8;
  static final short CDEV_INVALID    = 9;
  static final short CDEV_LONG       = 10;
  static final int CDEV_NTYPES = 11;
  static final int CDEV_SHORT = CDEV_INT16;
  static final int CDEV_INT = CDEV_INT32;

  static final short[] cdevToLocalType = {BYTE,SHORT,INT,INT,LONG,
					  FLOAT,DOUBLE,STRING,DOUBLE,
					  0,LONG};
  static final short[] localToCdevType = {CDEV_INVALID,CDEV_BYTE,CDEV_SHORT,
					  CDEV_INT,CDEV_LONG,CDEV_FLOAT,
					  CDEV_DOUBLE,CDEV_STRING};

  static final int MAXDIM = 10;

  Vector store;   // dynamic sized collection of DataEntry's

  //**************************************************************************
  //	Data ()
  /**
   * Construct an empty Data object.
   */
  public Data() {
    store = new Vector();
  }

  //**************************************************************************
  //	Data (xdrReader in)
  /**
   * Construct a Data object given the byte stream in an xdrReader object.
   * @param in - An xdrReader object that contains the byte stream.
   */
  public Data (xdrReader in) throws IOException {
    store = new Vector();
    xdrImport( in );
  }

  //**************************************************************************
  //	Data (Data d)
  /**
   * Shallow Copy Constructor.
   * @param original  The original Data object to replicate.
   */
  public Data (Data original)		// shallow copy
  {
    store = (Vector)(original.store).clone();
  }


  //**************************************************************************
  //	convertKey (String key)
  /**
   * Get the integer equivalent of a string key.
   * @param  key    The String key.
   * @return        The integer key.
   */
  public static int convertKey (String key) {
    return TagTable.convertKey(key);
  }

  //**************************************************************************
  //	convertKey (int key)
  /**
   * Get the string equivalent of an integer key.
   * @param  key    The integer key.
   * @return        The String key.
   */
  public static String convertKey (int key) {
    return TagTable.convertKey(key);
  }

  //**************************************************************************
  //	insertKeyPair (int key, String skey)
  /**
   * Insert a key pair into the table.
   * @param  key    The integer key.
   * @param  skey   The string key.
   */
  public static void insertKeyPair (int key, String skey) {
    TagTable.insert (key,skey);
  }

    /**
     * Returns a Class from an integer data type
     */
    public static Class getClass (int dataType) {
	return localTypeToClass[dataType];}


  //**************************************************************************
  //	getDataEntry (int key)
  /**
   * Return the <A href="cdev.data.DataEntry.html">DataEntry</A> with the
   * specified key, or null if not found.
   * @see     getDataEntry(int key)
   * @param   key   The integer key.
   * @return        A DataEntry object that matches the integer tag.
   */
  public DataEntry getDataEntry(int key) 
  throws IllegalArgumentException {
    for(int index=0; index < store.size(); index++) {
      DataEntry result = (DataEntry) store.elementAt(index);
      if (result.key == key) return result;
    }
    return null;
  }

  //**************************************************************************
  //	getDataEntry (String key)
  /**
   * Return the <A href="cdev.data.DataEntry.html">DataEntry</A> with the
   * specified key, or null if not found.
   * @see     getDataEntry(String key)
   * @param   tag   The String key.
   */
  public DataEntry getDataEntry(String key) {
    int ikey = TagTable.convertKey(key);
    return getDataEntry(ikey);
  }


  //**************************************************************************
  //	containsKey (int key)
  /** 
   * Test to see if the specified key is in the Data object.
   * @return true  If the key exists in this Data object.
   */
  public boolean containsKey(int key) {
    for(int index=0; index < store.size(); index++) {
      DataEntry result = (DataEntry) store.elementAt(index);
      if (result.key == key) return true;
    }
    return false;
  }

  //**************************************************************************
  //	containsKey (String key)
  /** 
   * Test to see if the specified key is in the Data object.
   * @return true  If the key exists in this Data object.
   */
  public boolean containsKey(String key) {
    return containsKey(convertKey(key));
  }

  //**************************************************************************
  //	toString ()
  /**
   * Convert the Data object to a formatted string.
   */
  public String toString() { return toString("contents: "); }

  //**************************************************************************
  //	toString (String title)
  /**
   * Convert the Data object to a string with a given title
   * preceeding the object.
   */
  public String toString(String title) {
    StringBuffer result = new StringBuffer();
    if (title != null)
      result.append(title).append("\n");
    for (int index=0; index<store.size(); index++) {
      DataEntry entry = (DataEntry) store.elementAt(index);
      result.append(entry.toString());
    }
  return result.toString();
  }

  //**************************************************************************
  //	addDataEntry (DataEntry entry)
  /**
   * Add a data entry to this container.
   */
  public void addDataEntry(DataEntry entry) {
    remove(entry.getKey());
    store.addElement(entry);
  }

  //**************************************************************************
  //	remove ()
  /**
   * Remove all <A href="cdev.data.DataEntry.html">DataEntry</A> objects from this 
   * Data object.
   */
  public void remove() {
    store.removeAllElements();
  }


  //**************************************************************************
  //	remove (int key)
  /**
   * Remove the data entry corresponding to a given key.
   * If the key is not found, do nothing.
   * @param key    Integer key.
   */
  public void remove(int key) {
    for (int index=0; index < store.size(); index++) {
      DataEntry de = (DataEntry) store.elementAt(index);
      if (de.key==key) {
        store.removeElementAt(index);
        break;
      }
    }
  }


  //**************************************************************************
  //	remove (String key)
  /**
   * Remove the data entry corresponding to a given key.
   * If the key is not found, do nothing.
   * @param key    String key.
   */
  //----------------------------------------------------------------------
  public void remove(String key) {
    int ikey = TagTable.convertKey(key);
    remove(ikey);
  }

// ****************************************************************************
// * insert:
  /**
   * Insert a scalar value into a Data object as a tagged data item. If the 
   * tag is in use, it is first removed!
   * @param  tag   The tag to associate with the data item.
   * @param  data  A byte data item.
   */
// ****************************************************************************
  public void insert(int tag, byte data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  //----------------------------------------------------------------------
  /**
   * Insert a scalar value into a Data object as a tagged data item. If the
   * tag is in use, it is first removed!
   * @param  tag   The tag to associate with the data item.
   * @param  data  A short data item.
   */
  public void insert(int tag, short data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  //----------------------------------------------------------------------
  /**
   * Insert a scalar value of type int into a Data object and give it 
   * the specified integer tag.
   * @param  tag   The tag to associate with the data item.
   * @param  data  An int data item to insert.
   */
  public void insert(int tag, int data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  //----------------------------------------------------------------------
  /**
   * Insert a scalar value of type long into a Data object and give it
   * the specified integer tag.  If the tag is used, the new item replaces
   * the old item.
   * @param  tag   The tag to associate with the data item.
   * @param  data  the data item to insert.
   */
  public void insert(int tag, long data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  //----------------------------------------------------------------------
  /**
   * Insert a scalar value of type float into a Data object and give it 
   * the specified integer tag.  If the tag is used, the new item replaces
   * the old item.
   * @param  tag   The tag to associate with the data item.
   * @param  data  the data item to insert.
   */
  public void insert(int tag, float data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  //----------------------------------------------------------------------
  /**
   * Insert a scalar value of type double into the Data object and give 
   * it the specified integer tag.  If the tag is used, the new item replaces
   * the old item.
   * @param  tag   The tag to associate with the data item.
   * @param  data  the data item to insert.
   */
  public void insert(int tag, double data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  //----------------------------------------------------------------------
  /**
   * Insert an object into the Data object and give it the specified integer 
   * tag.  If the tag is used, the new item replaces the old item.
   * @param  tag   The tag to associate with the data item.
   * @param  data  the data item to insert.
   */
  // Insertion of Strings, and arrays: 
  public void insert (int tag, Object data) {
    remove(tag);
    DataEntry entry = new DataEntry(tag,data);
    store.addElement(entry);
  }

  // Insertion by string tag:
  /**
   * Insert "data" into the this cdev.Data object associated with "tag".
   * @param tag      The string tag name to give the data
   * @param data     The byte data
   */
  public void insert (String tag, byte data) {
    //System.out.println ("Data::insert(string,byte): insert ["+data+"] into ["+tag+"]...");
    insert(convertKey(tag),data);
  }


  /**
   * Insert "data" into the this cdev.Data object associated with "tag".
   * @param tag      The string tag name to give the data
   * @param data     The short data
   */
  public void insert (String tag, short data) {
    //System.out.println ("Data::insert(string,short): insert ["+data+"] into ["+tag+"]...");
    insert(convertKey(tag),data);
  }


  /**
   * Insert "data" into the this cdev.Data object associated with "tag".
   * @param tag      The string tag name to give the data
   * @param data     The int data
   */
  public void insert (String tag, int data) {
    //System.out.println ("Data::insert(string,int): insert ["+data+"] into ["+tag+"]...");
    insert(convertKey(tag),data);
  }
  public void insert (String tag, long data) {
    insert(convertKey(tag),data);
  }
  public void insert (String tag, float data) {
    insert(convertKey(tag),data);
  }
  public void insert (String tag, double data) {
    insert(convertKey(tag),data);
  }
  public void insert (String tag, Object data) {
    insert(convertKey(tag),data);
  }

  //**************************************************************************
  //	changeKey (int oldkey, int newkey)
  /**
   * Change the key for an existing data entry.
   * If the old one can not be not found,
   * or the new tag already exists in this Data object,
   * an IllegalArgumentException is thrown.
   * @param   oldtag   The old tag number.
   * @param   newtag   The new tag number.   
   */
  public void changeKey(int oldkey, int newkey)
                        throws IllegalArgumentException 
  {
    for (int index=0; index < store.size(); index++) {
      DataEntry de = (DataEntry) store.elementAt(index);
      if (de.getKey()==oldkey) {
        if (containsKey(newkey)) {
	  throw new IllegalArgumentException("duplicate key" + newkey);
	}
	de.setKey(newkey);
        return;
      }
    }
    throw new IllegalArgumentException("key not found" + oldkey);
  }


  //**************************************************************************
  //	getValue (int key)
  /**
   *	Retrieves a data object specified by key from the specified DataEntry.
   *	If the key is not found, throws IllegalArgumentException.  
   *	@param	key	The key of the selected entry.
   *	@return		A reference to the contained data (array, Number, ...)
   */
  public Object getValue(int key) {
    DataEntry entry = getDataEntry(key);
    return entry.getValue();
  }

  //**************************************************************************
  //	getValue (String key)
  /**
   *	Retrieves a data object specified by key from the specified DataEntry.
   *	If the key is not found, throws IllegalArgumentException.
   *	@param	key	The key of the selected entry.
   *	@return		A reference to the contained data (array, Number, ...)
   */
  public Object getValue(String key) {
    DataEntry entry = getDataEntry(key);
    return entry.getValue();
  }

  //**************************************************************************
  //	elements ()
  /**
   *	Get an Enumeration of all contained DataEntry objects.
   */
  public Enumeration elements() {
    return store.elements();
  }


  //**************************************************************************
  //	xdrImport (xdrReader in)
  /**	Populate this Data object from a byte buffer which was written in XDR
   *	format.
   */
  public void xdrImport ( xdrReader bin ) throws IOException
  {
    int key  = 0;
    int type = 0;
    int dim  = 0;
    int elem = 0;    

    //    System.out.println ("inside xdrImport");
    //----- Read the byte cnt for this Data --------//
    // Note: This byte count is not used.  But must be
    //       read because the server sends a byte count
    //       with each Data export. 
    //--------------------------------------------------//
    bin.get_int();		// discard the total byte length

    //----- Get the number of DataEntry's ---//
    int cdevElementCnt = bin.get_int();
    //    System.out.println ("xdrImport: "+cdevElementCnt+" to import...");

    for (int i=0; i<cdevElementCnt; i++) {

      //----- Get header information -------------//
      key  = bin.get_int();
      type = bin.get_int();
      dim  = bin.get_int();
      elem = bin.get_int(); // This may not be useful since Java can't treat
                            // multi-dim arrays as 1D arrays.

      //System.out.println ( "Tag = "+tag+" :: type = "+type+
      //		   " :: dim = "+dim+" :: elemCnt = "+elem);

      //----- Get array dimensions -------------------//
      int[] Dim = null;          // necessary for readXdrArray
      if (dim != 0) {
        Dim = new int[dim];      // holding place for array dims
        for (int d=0; d<dim; d++) {       
	  bin.get_int();	// discard bounds offset (not supported)
          Dim[d] = bin.get_int();
        }
      }
      
      //----- Load the data  ---------------------//
      int DimCnt = 0;
      Object data = readXdrArray (dim, DimCnt, Dim, type, bin);
      
      //----- create the DataEntry object ----//
      DataEntry entry = new DataEntry(key, data);
      if (dim==0) entry.ndims = 0;

      //----- Add the entry into the Data ----//
      store.addElement(entry);

    } // repeat for each DataEntry element in the Data object

  }


  //**************************************************************************
  //	readXdrArray (,,,,)
  //    Recursive companion method for xdrImport to create a  
  //    datum or an n-dimensional data array.  Data is read from an
  //    xdrReader buffer.  
  //
  //    Caution:
  //    Recursion will not occur if numDim = 0. In this case, Dim[] will 
  //	be null which will throw an exception when Dim.length is checked.  
  //	This makes the xdrImport code shorter.
  private Object readXdrArray ( int numDim       // number of dimensions to expect
				,int D            // dimension of current iteration
				,int[] Dim        // the dimension array
				,int datatype     // the external data type
				,xdrReader bin  ) // the data buffer
    throws IOException
  { 
    Object result = null;

    //----- Recursively check for last dimension ---------------//
    if ( (numDim != 0) && (D < (Dim.length-1)) ) {     // not reached bottom
      int nextD = D + 1;
      Object[] array = new Object[Dim[D]];   // allocate the array pointers
      for (int i=0; i<Dim[D]; i++) {
        array[i] = readXdrArray (numDim, nextD, Dim, datatype, bin);
      }
      result = (Object) array;		// result has scope outside switch case

    //----- Load the data based on the type information -------------//
    } else { 
      switch (datatype)   {

        case CDEV_BYTE:
          if ( numDim == 0 ) {                // read single byte           
	    byte[] DATA = new byte[1];
	    DATA[0] = bin.get_byte();
            result = (Object)DATA;
          } else {                            // read byte array
	    byte[] DATA = new byte[Dim[D]];
            for (int k=0; k<Dim[D]; k++) 
              DATA[k] = bin.get_byte();
            result = (Object)DATA;
          }
          break;

        case CDEV_UINT16:
        case CDEV_INT16:
          if ( numDim == 0 ) {                // read single short
	    short[] DATA = new short[1];
	    DATA[0] = bin.get_short();
            result = (Object)DATA;
          } else {                            // read short array
	    short[] DATA = new short[Dim[D]];
            for (int k=0; k<Dim[D]; k++)
              DATA[k] = bin.get_short();
            result = (Object)DATA;
          }
          break;

	case CDEV_UINT32:
	case CDEV_INT32:
          if ( numDim == 0 ) {                // read single int
            int[] DATA = new int[1];
	    DATA[0] = bin.get_int();
            result = (Object)DATA;
          } else {                            // read int array
	    int[] DATA = new int[Dim[D]];
            for (int k=0; k<Dim[D]; k++) 
              DATA[k] = bin.get_int();
            result = (Object)DATA;
          }
          break;

        case CDEV_FLOAT:
          if ( numDim == 0 ) {                // read single float
            float[] DATA = new float[1];
	    DATA[0] = bin.get_float();
            result = (Object)DATA;
          } else {                            // read float array
	    float[] DATA = new float[Dim[D]];
            for (int k=0; k<Dim[D]; k++ ) 
              DATA[k] = bin.get_float();
            result = (Object)DATA;
          }
          break;

        case CDEV_DOUBLE:
          if ( numDim == 0 ) {                // read double
            double[] DATA = new double[1];
	    DATA[0] = bin.get_double();
            result = (Object)DATA;
          } else {                            // read double array
	    double[] DATA = new double[Dim[D]];
            for (int k=0; k<Dim[D]; k++) 
              DATA[k] = bin.get_double();
            result = (Object)DATA;
          }
          break;

        case CDEV_STRING:
          if ( numDim == 0 ) {                // read single string
            String DATA = bin.get_string();
            result = DATA;
          } else {                            // read string array
	    if (Dim[D]>MAXDIM)
	      System.out.println("xdr import String array, len="+Dim[D]);
	    String[] DATA = bin.get_string(Dim[D]);
            result = DATA;
          }
          break;

        case CDEV_TIMESTAMP:
          if ( numDim == 0 ) {                // convert TimeStamp to double
	    double dat = (double) bin.get_int();
	    dat += ((double) bin.get_int())/1000000000.; // nsec -> sec
            double[] DATA  = new double[1];
	    DATA[0] = dat;
            result = (Object)DATA;
          } else {
	    double[] DATA = new double[Dim[D]];
            for (int k=0; k<Dim[D]; k++ ) {
              DATA[k] = (double) bin.get_int();
	      DATA[k] += ((double) bin.get_int())/1000000000.;
	    }
	    result = (Object)DATA;
          }
          break;

	case CDEV_INVALID:
	  break;

        default:
	  break;
      }    // end of switch (datatype)
    }

    return (result);    // return the array as an object
  }


  //**************************************************************************
  //	xdrExport ()
  // *    Write this Data object into a buffer.  Buffer is written
  // *    in XDR format.  Most of this code should really be in DataEntry.
  // *    Warning:
  // *    The logic for determining the value of element cnt is in several
  // *    places and is quite tricky.
  // *       elem = 0               for scalar java native types
  // *       elem = element count   for arrays of java native types & String[]
  // *       elem = char count + 1  for single strings
  // public void xdrExport (xdrWriter bin)  
  // passing the writer in is more efficient
  // because the buffer does not have to be recopied!
  public byte[] xdrExport ()
  {
    ByteArrayOutputStream buffer = new ByteArrayOutputStream();  // create the buffer
    xdrWriter      bin = new xdrWriter(buffer);
    int            type = 0;                        // type of data
    int            elem = 0;                        // element count
                                                    // scalars have 0 in  count
    int            cdevElementCnt = store.size();
    DataEntry	   entry;                           // node to convert
    int            cnt;                             // temporary counter

    bin.put(cdevElementCnt);           // first put in the count
                                       // may need to check for empty vectors.

    //----- loop on index - faster than vector.isEmpty() call! ---------//
    for (int index=0; index<cdevElementCnt; index++)  {
      
      //----- setup pointers ------------------------//
      entry = (DataEntry)store.elementAt(index);
      Object  dat = entry.getValue();


      //----- get array info for packet header ------//
      if (entry.getDim() == 0)
	elem = 0;
      else
	elem = entry.nelems();

      //----- identify the data type ---------------//
      type = entry.getEnumeratedType();
     
      if (type == STRING) {       // String type
        if (entry.getDim() == 0) {
          String tempStr = ((String[]) dat)[0]; // recast as string to get length
          elem = tempStr.length() + 1;     // c++ strings need space for '\0'
        }
      }

      //----- pack the header data -----------------//

      bin.put(entry.getKey());          // pack tag
      int cdevType = localToCdevType[type];
      bin.put(cdevType);                // pack data type
      bin.put(entry.getDim());          // pack dimensions
      bin.put(elem);                    // pack num of elements

      //----- pack bounds --------------------------//
      // Skip if this DataEntry is not an array.  
      // Also compute the length for each dimension.
      //--------------------------------------------//
      if (entry.getDim() != 0) {         
        cnt = 0;
        int [] dims = entry.getDimensions();
        for (int i=0;i<dims.length; i++ ) {
     	  bin.put( (int)0  );           // for compatibility with C++ bounds  
          bin.put( dims[i] );
        }
      }

      //----- finally store the actual data ---------//
      //      System.out.println ("writing the data...");
      writeXdrArray (entry.getValue(), bin);

    }  // Repeat for each entry in the store vector.

    return ( buffer.toByteArray() );
  }


  //**************************************************************************
  //	writeXdrArray ()
  // Recursive method to write an n-dimentional array into an xdrWrite buffer.
  // Caution:
  // For native types, the recursion continues til the last dimension;  
  // however, for derived types, eg. Integer, Float..., the recursion goes
  // to the individual element. 
  void writeXdrArray (Object data, xdrWriter bin)
  { 
    //----- check to see if bottom of recursion reached -----//   

    if (data instanceof Object[])  {          // bottom not reached
      Object[] tdata = (Object[])data;
      for (int i=0; i<tdata.length; i++) {
	Object nextdata = tdata[i];           // peel off one index
	//        System.out.println ("xdrExport recurse...");
        writeXdrArray (nextdata, bin);        // recursive call
      }

    } else {                                  // bottomed out in the recursion
      //----- handle native types arrays AND String ------------//
      // Native types have 1 less recursion step 
      // so need to loop.  Unfortunately code for
      // loop has to be repeated to get correct function 
      // (overloaded).
      if (data instanceof byte[])         {       // byte[]
	byte[] DATA = (byte[]) data;
	for (int k=0; k<DATA.length; k++)
	  bin.put(DATA[k]);              
      } else if (data instanceof char[])  {        // char[] 
	char[] DATA = (char[]) data;
	for (int k=0; k<DATA.length; k++)
	  bin.put(DATA[k]);              
      } else if (data instanceof short[]) {         // short[]
	short[] DATA = (short[]) data;
	for (int k=0; k<DATA.length; k++)
	  bin.put(DATA[k]);              
      } else if (data instanceof int[])   {        // int[]  
	int[] DATA = (int[]) data;
	//          System.out.println ("writing int[]...");
	for (int k=0; k<DATA.length; k++)  {
	  bin.put(DATA[k]);
	  //	    System.out.println ("DATA["+k+"]: "+DATA[k]);
	}
      } else if (data instanceof long[])  {        // long[]  
	long[] DATA = (long[]) data;
	for (int k=0; k<DATA.length; k++)
	  bin.put(DATA[k]);              
      } else if (data instanceof float[]) {        // float[] 
	float[] DATA = (float[]) data;
	for (int k=0; k<DATA.length; k++)
	  bin.put(DATA[k]);              
      } else if (data instanceof double[]){        // double[] 
	double[] DATA = (double[]) data;
	for (int k=0; k<DATA.length; k++)
	  bin.put(DATA[k]);              
      } else if (data instanceof String) {
 	bin.put_string((String)data);
      }
    }
  }

  //**************************************************************************
  //	getSize()
  /**
   * Return the number of <A href="cdev.data.DataEntry.html">DataEntry</A> objects
   * in this Data object.
   */
  public int getSize ()  {    return store.size();  }


  //**************************************************************************
  //	equals ()
  /**
   * Perform a deep compare of 2 Data object to determine if they have
   * the same data.  The comparison is performed by converting the data
   * objects first to a byte streams then comparing the byte streams.
   */
  public boolean equals (Data other)
  {
    boolean result = false;

    if ( store.size() == other.getSize() )  {
      byte[] otherBuf = other.xdrExport();
      byte[] thisBuf = this.xdrExport();
      if (thisBuf.length == otherBuf.length ) {
        int i = 0;
        result = true;
        while ( (result) && (i < thisBuf.length) )  {
          result = (thisBuf[i] == otherBuf[i])?true:false;
          i++;
        }   // end while
      }  // end bufsize check
      thisBuf = null;    // free the memory
      otherBuf = null;
    }  // end num entry check
    return result;
  }


    /**
     * clone():
     * performs a deep copy of a Data Object
     * @return  the clone Object
     */

  // Author: M.C
     

  public synchronized Object clone() throws CloneNotSupportedException {
    
    Data d = (Data) super.clone(); //cloning the Data structure
    d.store = (Vector) store.clone(); //cloning the Vector structure
    for (int i=0;i<store.size();i++) { //cloning DataEntry in the vector
      (d.store).setElementAt(((DataEntry)store.elementAt(i)).clone(),i);
    } 
    return d; 
  }


  /**
   * Performs a copy of this Data into another destination Data.
   * DataEntries from this Data are copied into the destination Data.
   * If the destination Data contains DataEntries with the same key, same
   * type and same dimensions as this Data, the DataEntries of this Data
   * will be copied into the destination Data ones, if not, they will be cloned.
   * The DataEntries contained in the destination Data whose key is not
   * the key of  any DataEntry of this Data are not affected.
   * @param dst  The destination Data. Cannot be null. 
   * @see   DataEntry.smartCompareCopy(DataEntry dst)
   */

  // Author: M.C

  public synchronized void copyInto(Data dst) throws CloneNotSupportedException {


    DataEntry dataEntrySrc = null;
    int srcKey = 0;

  //if Data destination exists
    if (dst != null) {
    // for each DataEntry of Data source 
      for (int index=0; index<this.store.size();index++) {
      //try to find one with the same key in Data destination
	dataEntrySrc = (DataEntry)this.store.elementAt(index);
	srcKey = dataEntrySrc.getKey();
	DataEntry dataEntryDst = dst.getDataEntry(srcKey); // DataEntryDst is null if key not found
  
        // perform a smart copy (copy or clone depends on existence and matching field of source DataEntry and destination
	DataEntry dataEntryTemp = dataEntrySrc.smartCompareCopy(dataEntryDst);
	if (dataEntryTemp != null) {  // if DataEntries were not the same
	  if (dataEntryTemp.equals(dataEntryDst) == false) { //if clone has been used
	  dst.addDataEntry(dataEntryTemp); //add the new DataEntry to destination Data. 
	  }
	}
      } //end of for each...
	
    }

  }

}


