//---------------------------------------------------------------------------
// Copyright (c) 1991,1992 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description: xdrWriter.java
//      Class to translate Java native types into buffered XDR stream.
//      User responsible for calling flush() to send data out if writing
//      directly to buffered socket.
//  
//  Issues:
//      1) Unsigned types not implemented.  But byte writes allowed.  Is this
//         adaquate?
//
//  Author: Cuong Quach
//
//----------------------------------------------------------------------------
//
// Revision History:
//
//  9-5-96 :  Created CQ.
//  9-12-96:  Added overload for "put" methods.  CQ.
//  9-20-96:  Added "finalize" method to close socket. CQ.
//  10-6-96:  Added implement queing, modified flush to accommodate. CQ.
//  11-11-96: Make the xdr simple type writting methods final.
//  12-3-98:  Removed CLIP specific code. CW.
//
//----------------------------------------------------------------------------

package cdev.data;

import java.io.*;
import java.net.*;
import java.util.*;

public class xdrWriter
{
  private  DataOutputStream      xdrOut = null;

//-------------------------------------------------------------//
// xdrWriter constructor:
// Creates a writer connected to an output stream.
//-------------------------------------------------------------//
  public xdrWriter (OutputStream output) 
  { 
    try {
      xdrOut = new DataOutputStream(output);
    } catch (Exception e) {  e.printStackTrace();  }
  }

//-------------------------------------------------------------//
// Returns the byte count written to buffer so far.
//-------------------------------------------------------------//
  public final int size()
  {
    return ( xdrOut.size() );
  }


//-------------------------------------------------------------//
//  Write Characters
//-------------------------------------------------------------//
  public final int put_char (char x)
  {
    try {
      xdrOut.writeByte(0);          // pad 3 bytes
      xdrOut.writeByte(0);
      xdrOut.writeByte(0);
      xdrOut.writeByte((byte) x);   // low byte of Unicode char
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  }

  public int put (char x)  { return ( put_char(x) );  }


//-------------------------------------------------------------//
//  Write Integers
//-------------------------------------------------------------//
  public final int put_int(int x)
  {
    try {
      xdrOut.writeInt(x);
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  } 

  public final int put (int x)      { return ( put_int(x) ); }
  public final int put (Integer x)  { return ( put_int(x.intValue()) ); }


//-------------------------------------------------------------//
//  Write Shorts
//-------------------------------------------------------------//
  public final int put_short(short x)
  {
    try {
      xdrOut.writeByte(0);
      xdrOut.writeByte(0);
      xdrOut.writeShort(x);
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  } 

  public final int put (short x)  { return ( put_short(x) );  }


//-------------------------------------------------------------//
// Write Long
//-------------------------------------------------------------//
  public final int put_long (long x)
  {
    try {
      xdrOut.writeInt((int) x);   // on Solaris long = int = 4bytes
//      xdrOut.writeLong(x);      // may need this in other OS
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  }

  public final int put (long x)  {  return ( put_long(x) );  }
  public final int put (Long x)  {  return ( put_long(x.longValue()) );  }


//-------------------------------------------------------------//
// Write Floats
//-------------------------------------------------------------//
  public final int put_float (float x)
  {
    try {
      xdrOut.writeFloat (x);
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  }

  public final int put (float x)  {  return ( put_float(x) );  }
  public final int put (Float x)  {  return ( put_float(x.floatValue()) );  }


//-------------------------------------------------------------//
// Write Doubles
//-------------------------------------------------------------//
  public final int put_double (double x)
  {
    try {
      xdrOut.writeDouble (x);
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  }

  public final int put (double x)  { return ( put_double(x) );  }
  public final int put (Double x)  { return ( put_double(x.doubleValue()) );  }


//-------------------------------------------------------------//
// Write Strings
//-------------------------------------------------------------//
  public final int put_string (String x)
  {
    int len = x.length();
    
    try {

      //-----------------------------------------------------//
      // writeUTF will write a 2-byte count followed by one
      // byte characters.  Must first pad 2 bytes to make it 
      // a 4-byte count.
      //-----------------------------------------------------//
      xdrOut.writeShort((short) 0);
      xdrOut.writeUTF (x);              // write UTF-8 string
      if ( (len%4) > 0 ) {	        // now fill with pad bytes
        int padlen = 4 - (len % 4);
	for (int i=0; i<padlen; i++) {
          xdrOut.writeByte(0);	  
	}
      }
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;
  }
  
  public final int put (String x)  {  return ( put_string(x) );   }
  
  public final int put (String[] x)   // overloaded string array put
  { 
    int status = 0;
    put_int(x.length);                    // load the count
    for (int i=0; i<x.length; i++) {      // load each string
      status = put_string (x[i]);
    }
    return (status);
  }


//-------------------------------------------------------------//
// Write bytes & opaque
//-------------------------------------------------------------//
  public final int put_opaque (byte[] b)
  {
    if (b == null) return -1;    // trying to write a null array
    try {
      xdrOut.writeInt(b.length);             // put the length first
      xdrOut.write (b, 0, (b.length) );      // put the data bytes
//      xdrOut.write (b, 0, (b.length-1) );    // put the data bytes
      if ( (b.length%4) > 0 ) {	             // now fill with pad bytes
        int padlen = 4 - (b.length % 4);
	for (int i=0; i<padlen; i++) {
          xdrOut.writeByte(0);	  
	}
      }
    } catch (Exception e)  {
      System.out.println(e);
      return (-1);      
    }
    return 0;
  }

  public final int put_byte(byte x)
  {
    try {
      xdrOut.writeByte((int)x);
    } catch (Exception e) {
      System.out.println(e);
      return (-1);
    }
    return 0;    
  }

  public final int put       (byte x)     {  return put_byte(x);    }
  public final int put       (byte[] x)   {  return put_opaque(x);  }


//-------------------------------------------------------------//
// flush:
// This function actually writes data out the socket if a socket
// is connected.  This must be called to send data out a socket.
//
//-------------------------------------------------------------//
  public final int flush ()   
  {
    try {
      xdrOut.flush();
    } catch (Exception e) {
      System.out.println(e);
      return (-1);      
    }
    return 0;
  }

//-------------------------------------------------------------//
//  Special method involked just before the memory manager 
//  recycles the memory for this object.  If a socket is attached
//  the socket should be released because the memory manager
//  does not do this automatically.
//-------------------------------------------------------------//
  protected void finalize () throws Throwable
  {
    try { xdrOut.close(); }
    catch (IOException e) {    }    // no error recovery
    finally {super.finalize();}
  }

}  // end of xdrWriter class
