package cdev.dir;

import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Interface for the meaning of bitgroups in a bitpattern.
 *
 * By far the most common bitpattern will be an enumeration. If you are sure
 * this is the only thing you will encounter, then methods
 * <UL>
 * <LI> int <a href="#convertToValue(java.lang.String)">
 *      convertToValue(String label)</a> -
 *      convert a label to an integer value
 * <LI> String <a href="#convertToString(int)">
 *      convertToString(int value)</a> -
 *      convert an integer value to a label
 * <LI> String[] <a href="#getLabels(int)">getLabels(0)</a> -
 *      get an array of defined labels
 * </UL>
 * are all you need and you can skip the rest of this description.
 *
 * But the real world can be more complex and lot of information is
 * in packed form such as status words coming from device or interface
 * registers. A generic program can make sence of these
 * bitpatterns with the help of this class.
 * <P>
 * A bitpattern is an integer data word wich is seen as a composite structure.
 * A bitpattern can contain one or more consecutive bitgroups and
 * a bitgroup can contain one or more consecutive bits. A bitgroup is
 * identified by firstBit and ends before the firstBit of the next group,
 * if any. The value of a bitpattern is the value of the data word seen
 * as an integer.
 * <P>
 * It must be possible to mask unimportant bits in the bitpattern
 * (this treats them as if they were 0) but this is done transparantly by
 * the implementation.
 * <P>
 * A bitgroup is defined by one or more records composed of:
 * <UL>
 * <LI><B>firstBit</B> : the first bit in the bitgroup (from 0)
 * <LI><B>groupValue</B> : reference value for the bitgroup seen as an
 * integer, with firstBit having a weight of 1.
 * <LI><B>label</B> : string corresponding to groupValue.
 * Labels should be unique for the BitPatternDefinition.
 * </UL>
 * A bitgroup is discrete if it has corresponding defined pairs of
 * groupValue and label.
 * An exclusive group of bits is just a special case of a discrete group.
 * <P>
 * A bitgroup is continuous if its bits are read as an integer value.
 * A continuous group is defined by a single record with reference value=-1
 * (so that it will match no possible actual value)
 * and the label indicates the meaning of the value. The label seen in
 * conversions (on input or output) is in fact a concatenation: 'label=value',
 * for instance: 'widget position=15'. This makes decoding of packed
 * values easier.
 * <P>
 * A BitPatternDefinition object can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getBitPatternDefinition">
 * DirectoryService#getBitPatternDefinition(patternName)</a>.
 * The patternName must be unique for the implementation.
 * <BR>Implementation with a local cache is recommended for efficiency.
 * <P>
 * Note: in the bit description, describe only interesting situations: if
 * a bit=1 indicates a fault, then there should be a matching label, but
 * an absence of the fault (here bit=0) should have no matching label.
 * <P>
 * General documentation in <a href="DirectoryService.html">DirectoryService</a> .
 *
 * @author   J.Cuperus
 * @version  1.0
 *
 * @see DirectoryService
 * @see DeviceProperty
 * @see DeviceMessage
 * @see cdev.dir.rdbms.RdbmsBitPatternDefinition
 */
public interface BitPatternDefinition extends DirectoryItem {

/**
 * Default separator (',') for concatenated labels.
 */
public final char DEFAULT_SEPARATOR = ',';

//---------------------------------------------------------------------------

/**
 * Converts a bitpattern to an array of group values.
 * One value per bitgroup is returned, in the sequence of the bitgroups.
 *
 * @param value the integer to be converted
 * @return array of group values
 */
public long[] convertToGroupValues(long value);

//---------------------------------------------------------------------------

/**
 * Converts a bitpattern to a DeviceMessage.
 * Only the message with the highest level is returned.
 *
 * @param value the integer to be converted
 * @return most relevant DeviceMessage
 * @exception NoSuchDataException -
 */
public DeviceMessage convertToMessage(long value)
       throws NoSuchDataException;

//---------------------------------------------------------------------------

/**
 * Converts a bitpattern to a label (for an enumeration), or a
 * concatenation of labels for a more complex bitpattern.
 * The separator is DEFAULT_SEPARATOR.
 *
 * @param value the integer to be converted to string
 * @return the label or concatenated labels corresponding to the value
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public String convertToString(long value)
       throws NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Converts a bitpattern to a string of concatenated labels.
 * Returns labels, separated by <I>separator</I> in the sequence of the
 * bitgroups but bitgroups which have no corresponding label are skipped.
 *
 * @param value the integer to be converted
 * @return concatenation of one or more labels
 * @exception NoSuchDataException    -
 */
public String convertToString(long value, char separator)
       throws NoSuchDataException;

//---------------------------------------------------------------------------

/**
 * Converts a label, or a concatenation of labels separated by
 * DEFAULT_SEPARATOR, to a value.
 *
 * @param label the string to be converted to integer
 * @return the value of the label
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public long convertToValue(String label)
       throws NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Converts a concatenation of labels, with given separator, to a value.
 * The labels may be in any order.
 * For each label in concatenation, bitgroups are searched until the label is
 * found and its groupvalue is packed into the value.
 * The method fails if any of the labels is not found.
 *
 * @param labels the strings to be be packed in the bitpattern
 * @return the value of the bitpattern
 * @exception NoSuchDataException -
 * @exception BadQueryException   -
 */
public long convertToValue(String labelConcatenation, char separator)
       throws BadQueryException, NoSuchDataException;

//---------------------------------------------------------------------------

/**
 * Gets array of firstBits for bitgroups.
 *
 * @return array of firstBits
 */
public int[] getFirstBits();

//---------------------------------------------------------------------------

/**
 * Gets array of defined messages for bitgroup starting at firstBit
 *
 * @param  firstBit the first bit of the bitgroup
 * @return array of messages
 * @exception BadQueryException if bitgroup is not defined
 */
public DeviceMessage[] getMessages(int firstBit)
       throws BadQueryException;

//---------------------------------------------------------------------------

/**
 * Returns true if this bitpattern definition is a simple enumeration
 *
 * @return true if <I>this</I> is an enumeration
 */
public boolean isEnumeration();

//---------------------------------------------------------------------------

/**
 * Return true if this bitpattern 'contains' the label
 *
 * @param  value the value of the bitpattern seen as an integer
 * @param  label the presence of this label must be checked
 * @return true if this 'contains' the label
 */
public boolean isThisLabel(long value, String label);

//---------------------------------------------------------------------------

}
