package cdev.dir;

import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Interface DeviceData for fixed data about accelerator devices.
 * Included is information about related devices.
 * <P>
 * Some of the information for connecting to the device may
 * be site-specific. The site-specific variable names can be obtained with
 * <a href="#getSpecialVariableNames()">
 * getSpecialvariableNames</a> and their instance values with
 * <a href="#getSpecialValue(java.lang.String)">
 * getSpecialValue(variableName)</a>.
 * <P>
 * A DeviceData object can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceData(java.lang.String)">
 * DirectoryService#getDeviceData(deviceName)</a>.
 * <BR>An array of DeviceData objects in a devicelist can be obtained with
 * <a href="cdev.dir.DeviceList.html#getDeviceData()">
 * DirectoryService#getDeviceData</a>.
 * <P>
 * General documentation in <a href="DirectoryService.html">DirectoryService</a> .
 *
 * @author   J.Cuperus
 * @version  1.0
 *
 * @see DeviceData
 * @see DeviceClass
 * @see DeviceList
 */
public interface DeviceData extends DirectoryItem {

//---------------------------------------------------------------------------

/**
 * Gets the name of the device class.
 */
public String getClassName();

//---------------------------------------------------------------------------

/**
 * Gets the device class object.
 *
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 * @exception TooManyValuesException -
 */
public DeviceClass getDeviceClass()
       throws NoSuchDataException, BadConnectionException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Gets array of names of related devices for the role.
 *
 * @param  role  the name of the role (may have prefix ~ for passive role).
 * @return array of names of related devices; NoSuchDataException if none
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 */
public String[] getRelatedDevices(String role)
       throws NoSuchDataException, BadConnectionException;

//---------------------------------------------------------------------------

/**
 * Gets array of role names (prefix "~" for inverse role).
 * roleName is the active role of the device for a related device.
 * ~roleName is when a related device has the active role for this device.
 *
 * @return array of role names (prefix "~" for inverse role);
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 */
public String[] getRoles()
       throws NoSuchDataException, BadConnectionException;

//---------------------------------------------------------------------------

/**
 * Gets value for an implementation-specific variable as a string.
 * It is up to the application to know the type of the variable and do the
 * proper conversion.
 *
 * @param  SpecialVariableName the name of an application-specific variable.
 * @return the value of the variable as a string
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 */
public String getSpecialValue(String specialVariableName)
       throws NoSuchDataException, BadConnectionException;

//---------------------------------------------------------------------------

/**
 * Gets the names of implementation-specific variables.
 * These variables often indicate how to access the device. They may be
 * needed by the device I/O implementation.
 *
 * @return array of variable names; NoSuchDataException if none
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 */
public String[] getSpecialVariableNames()
       throws NoSuchDataException, BadConnectionException;

//---------------------------------------------------------------------------

/**
 * Returns true if this device has the role (prefix "~" for inverse role).
 * roleName is the active role of the device for a related device.
 * ~roleName is when a related device has the active role for this device.
 *
 * @param  role  the name of the role (may have prefix ~ for passive role).
 * @return true if this device has the role
 * @exception BadConnectionException -
 */
public boolean isDefined(String role)
       throws BadConnectionException;

//---------------------------------------------------------------------------

/**
 * Return true when the device depends on the cycle type.
 */
public boolean needCycleType();

//---------------------------------------------------------------------------

}
