package cdev.dir;

import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Interface DeviceList for a list of DeviceData for generic programs.
 * This list tells the application program what devices it should work on.
 * <BR>The DeviceData objects in the list are devided in groups identified
 * by a sequential number starting from 0.
 * <P>
 * A DeviceList object is identified by a name, which must be
 * unique for the implementation, or by a query.
 * <BR>An array of all devicelist names can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceListNames()">
 * DirectoryService#getDeviceListNames()</a>.
 * <BR>An array of selected devicelist names can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceListNames(java.lang.String)">
 * DirectoryService#getDeviceListNames(query)</a>.
 * <P>
 * A 'hand-made' DeviceList object can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceListByName(java.lang.String)">
 * DirectoryService#getDeviceListByName(listName)</a>.
 * <BR>An 'on-the-fly' DeviceList object can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceListByQuery(java.lang.String)">
 * DirectoryService#getDeviceListByQuery(query)</a>.
 * <P>
 * General documentation in <a href="DirectoryServiceDoc.html">DirectoryService</a> .
 *
 * @author   J.Cuperus
 * @version  1.0
 *
 * @see DeviceData
 * @see DeviceClass
 */
public interface DeviceList extends DirectoryItem {

//---------------------------------------------------------------------------

/**
 * Gets the number of devices in the device list.
 *
 * @return the device count
 */
public int getDeviceCount();

//---------------------------------------------------------------------------

/**
 * Gets the number of devices in group n of the device list.
 *
 * @param n group number (from 0)
 * @return device count (BadQueryException if not in [0..getGroupCount()-1])
 * @exception BadQueryException -
 */
public int getDeviceCount(int n)
       throws BadQueryException;

//---------------------------------------------------------------------------

/**
 * Gets the number of groups in the device list.
 *
 * @return the number of device groups
 */
public int getGroupCount();

//---------------------------------------------------------------------------

/**
 * Gets the DeviceData array.
 *
 * @return the DeviceData object array
 */
public DeviceData[] getDeviceData();

//---------------------------------------------------------------------------

/**
 * Gets the array of distinct class names in the device list.
 *
 * @return array of class names
 */
public String[] getClassNames();

//---------------------------------------------------------------------------

/**
 * Gets the array of device names in the device list.
 *
 * @return array of device names
 */
public String[] getDeviceNames();

//---------------------------------------------------------------------------

/**
 * Gets the array of DeviceData objects for group n.
 *
 * @param n  the group sequence number [0..getGroupCount()-1]
 * @return array of DeviceData objects (BadQueryException if n out of range)
 * @exception BadQueryException -
 */
public DeviceData[] getDeviceData(int n)
       throws BadQueryException;

//---------------------------------------------------------------------------

/**
 * Gets the array of device names for group n.
 *
 * @param n  the group sequence number [0..getGroupCount()-1]
 * @return array of device names (BadQueryException if n out of range)
 * @exception BadQueryException -
 */
public String[] getDeviceNames(int n)
       throws BadQueryException;

//---------------------------------------------------------------------------

/**
 * Gets the array of distinct class names for group n.
 *
 * @param n  the group sequence number [0..getGroupCount()-1]
 * @return array of class names (BadQueryException if n out of range)
 * @exception BadQueryException -
 */
public String[] getClassNames(int n)
       throws BadQueryException;

//---------------------------------------------------------------------------

}
