/* 
 * GatewayDirService.java
 *
 * Change History
 *  29-Aug-00 CW First version
 */

package cdev.dir.gateway;

import cdev.*;
import cdev.data.*;
import cdev.dir.*;

import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * For device class data and references to DeviceProperty objects.
 * <P>
 * <P>You can obtain lists of property names
 * defined for the device class with
 * <a href="#getPropertyNames()">
 *    getPropertyNames()</a> and
 * <a href="#getPropertyNames(java.lang.String)">
 *    getPropertyNames(listName)</a>.
 * You can obtain the DeviceProperty object itself with
 * <a href="#getDeviceProperty(java.lang.String)">
 * getDeviceProperty(propertyName)</a> Note: Not implemented in this version.
 * <P>
 * A DeviceClass object is identified by its name which must be
 * unique for the implementation.
 * <BR>An array of all deviceclass names can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceClassNames()">
 * DirectoryService#getDeviceClassNames()</a>.
 * <BR>A subset array of deviceclass names can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceClassNames(java.lang.String)">
 * DirectoryService#getDeviceClassNames(query)</a>.
 * <P>
 * A DeviceClass object can be obtained with
 * <a href="cdev.dir.DirectoryService.html#getDeviceClass(java.lang.String)">
 * DirectoryService#getDeviceClass(className)</a>.
 * <P>
 * General documentation in <a href="DirectoryService.html">DirectoryService</a> .
 *
 * @author   Chip Watson
 * @version  1.0
 *
 * @see DeviceData
 * @see DeviceProperty
 * @see DeviceList
 */
class GatewayDeviceClass implements DeviceClass {
  Device directory; // reference to the cdevDirectory (shared)
  String className; // name of the device class this represents

  GatewayDeviceClass(Device dir, String cname) {
    directory = dir;
    className = cname;
  }

  /**
   * Gets a one-line description of this directory object.
   * Mainly documentation for display to users but, in some cases like in
   * ErrorData it can convey essential information.
   *
   * @return    the description of the directory object (null if none)
   */
  public String getDescription()
  {
    return "DeviceClass "+className;
  }

  //---------------------------------------------------------------------------

  /**
   * Gets the name of the service which supplied the information.
   * In a mixed system, this identifies the subsystem that provides the
   * data. This information may also be used by the equipment access
   * classes to select the I/O implementation.
   *
   * @return    the name of the service that supplied the information
   */
  public String getServiceName()
  {
    return "cdevDirectory (gateway)";
  }

  //---------------------------------------------------------------------------

  /**
   * A multiline string with information about the object.
   * If 'all' is true, more complete information may be returned.
   *
   * @param all true if you want complete information
   * @return    the multiline information string
   */
  public String getInfo(boolean all)
  {
    return getDescription();
  }

  //---------------------------------------------------------------------------

  /**
   * Gets minimum number of lines to display the getInfo string.
   *
   * @return    the number of lines in the information string.
   */
  public int getInfoLines() {return 1;}

  //---------------------------------------------------------------------------

  /**
   * Gets minimum line width, in characters, to display the getInfo string.
   *
   * @return    the number of characters per line
   */
  public int getInfoWidth() {return 40;}

  //---------------------------------------------------------------------------

  /**
   * Gets the name of the directory object.
   *
   * @return    the name of the directory object
   */
  public String getName() {return className;}

  //---------------------------------------------------------------------------

  /**
   * Gets the identifier as a representation of the directory object.
   * This may be the name, if it is unique for the class, or a concatenation
   * of names like 'className/propertyName' which is unique.
   *
   * @return    the representation string for the directory object
   */
  public String toString() {return className;}
  //---------------------------------------------------------------------------

  /**
   * Gets DeviceProperty object, given the property name.
   *
   * @param propertyName the name of the property
   * @return DeviceProperty object
   * @exception NoSuchDataException    -
   * @exception BadConnectionException -
   * @exception TooManyValuesException -
   */
  public DeviceProperty getDeviceProperty(String propertyName)
    throws NoSuchDataException, BadConnectionException, TooManyValuesException 
  {
    throw new NoSuchDataException(propertyName);
  }
  

  //---------------------------------------------------------------------------

  /**
   * Gets array of all propertylist names for the deviceclass.
   * Array is ordered by list name.
   *
   * @return array of propertylist names
   * @exception NoSuchDataException    -
   * @exception BadConnectionException -
   */
  public String[] getPropertyListNames()
    throws NoSuchDataException, BadConnectionException
  {
    return null;
  }

  //---------------------------------------------------------------------------

  /**
   * Gets array of all property names for the deviceclass.
   * Array is ordered by property name.
   *
   * @return array of property names
   * @exception NoSuchDataException    -
   * @exception BadConnectionException -
   */
  public String[] getPropertyNames()
    throws NoSuchDataException, BadConnectionException
  {
    Data data = new Data();
    Data result = new Data();

    try {
      data.insert("class",className);
      DeviceError err = directory.send("queryAttributes",data,result);
      if (err == null)
	return result.getDataEntry("value").stringArray();
      else
	throw new NoSuchDataException(err.toString());
    } catch (IOException e) {
      throw new BadConnectionException(e.toString());
    } catch (Exception e) {
      throw new NoSuchDataException();
    }
  } 

//---------------------------------------------------------------------------

/**
 * Gets a ordered subset of the property names of a class.
 *
 * @param listName the name for the sublist of properties for the device class.
 * @return array of property names
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 */
public String[] getPropertyNames(String listName)
       throws NoSuchDataException, BadConnectionException
  {
    throw new NoSuchDataException("not implemented");
  }


  //---------------------------------------------------------------------------

}
