/* 
 * GatewayDirService.java
 *
 * Change History
 *  28-Aug-00 CW Added support for getting device class names
 */

package cdev.dir.gateway;

import cdev.*;
import cdev.data.*;
import cdev.dir.*;

import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * This class implements (a subset of) DirectoryService using the cdevGateway
 * and the device cdevDirectory.
 * <P>
 * As a first step you must get the implemented directory service
 * with a statement like:
 * <P>
 * DirectoryService dir = cdev.Config.getDirectoryService();
 * <P>
 * <P>
 * Several methods require a 'query' string for making a selection of
 * object identifiers. 
 * <P>
 * General documentation in <a href="DirectoryServiceDoc.html">DirectoryServiceDoc</a> .
 *
 * Note: only one DirectoryServiceImplementation is installed for a site.
 *       This implementation may however address <a href="SubServicesDoc.html"> 
 *	 multiple DirectoryItem implementations</a>
 *
 * @author   Chip Watson
 * @version  1.1
 *
 * @see DeviceData
 * @see DeviceClass
 * @see DeviceList
 * @see DeviceProperty
 * @see DeviceMessage
 * @see BitPatternDefinition
 */
public class GatewayDirService extends DirectoryService {

  private static GatewayDirService singleton = null;

  private Device directory;

  //===========================================================================
  // Static Methods
  //===========================================================================

  /**
   * Gets unique instantiation of the directory service implementation.
   * This static method must be overridden in the implementation.
   *
   * @exception BadConnectionException -
   */
  public static DirectoryService getService() throws BadConnectionException
  {
    try {
      if (singleton == null) singleton = new GatewayDirService();
      return singleton;
    } catch (Exception e) {throw new BadConnectionException(e.toString());}
  }

  //===========================================================================
  // Constructor
  //===========================================================================

  /**
   * Default constructor, never called directly.
   */
  protected GatewayDirService() throws IOException
  {
    directory = new Device("cdevDirectory");
  }


  //===========================================================================
  // DeviceClass
  //===========================================================================

  /**
   * Gets DeviceClass object when given the class name.
   *
   * @param className  the name of the device class
   * @return the DeviceClass object
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException -
   */
  public DeviceClass getDeviceClass(String className)
    throws BadConnectionException, NoSuchDataException, TooManyValuesException
  {
    return new GatewayDeviceClass(directory,className);
  }

  //---------------------------------------------------------------------------

  /**
   * Gets an array of all device class names.
   *
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   */
  public String[] getDeviceClassNames()
    throws BadConnectionException, NoSuchDataException
  {
    try {
      return getDeviceClassNames("");
    } catch (BadQueryException e) {
      throw new BadConnectionException(e.toString());
    }
  }

  //---------------------------------------------------------------------------

  /**
   * Gets an array of device class names from a query string.
   *
   * @param query  a selection statement depending on the implementation
   * @return array of device class names
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception BadQueryException      -
   */
  public String[] getDeviceClassNames(String query)
    throws BadConnectionException, NoSuchDataException, BadQueryException
  {
    Data data = new Data();
    Data result = new Data();

    try {
      data.insert("class",query);
      DeviceError err = directory.send("queryClasses",data,result);
      if (err == null)
	return result.getDataEntry("value").stringArray();
      else
	throw new NoSuchDataException(err.toString());
    } catch (IOException e) {
      throw new BadConnectionException(e.toString());
    } catch (Exception e) {
      throw new NoSuchDataException();
    }
  }


  //===========================================================================
  // DeviceData
  //===========================================================================

  /**
   * Gets the DeviceData object from the device name.
   *
   * @return the DeviceData object.
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException -
   */
  public DeviceData getDeviceData(String deviceName)
    throws NoSuchDataException
  { 
    throw new NoSuchDataException(deviceName);
  }

  //---------------------------------------------------------------------------

  /**
   * Gets device names from a query string.
   *
   * @param query  a selection statement depending on the implementation
   * @return array of device names
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception BadQueryException      -
   */
  public String[] getDeviceNames(String query)
    throws BadConnectionException, NoSuchDataException, BadQueryException
  {
    try {
      Data data = parse(query);
      Data result = new Data();
      DeviceError err = directory.send("query",data,result);
      if (err == null)
	return result.getDataEntry("value").stringArray();
      else
	throw new NoSuchDataException(err.toString());
    } catch (IOException e) {
      throw new BadConnectionException(e.toString());
    } catch (Exception e) {
      throw new NoSuchDataException();
    }
  }

  //===========================================================================
  // DeviceList
  //===========================================================================

  /**
   * Gets array of all devicelist names.
   *
   * @return array of device list names
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   */
  public String[] getDeviceListNames()
    throws BadConnectionException, NoSuchDataException
  {
    return null;
  }

  //---------------------------------------------------------------------------

  /**
   * Gets array of devicelist names from a database query.
   *
   * @return array of device list names
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception BadQueryException      -
   */
  public String[] getDeviceListNames(String query)
    throws BadConnectionException, NoSuchDataException, BadQueryException
  {
    return null;
  }

  //---------------------------------------------------------------------------

  /**
   * Gets DeviceList object from the name of a list.
   *
   * @param listName   the name of the device list
   * @return DeviceList object
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException -
   */
  public DeviceList getDeviceListByName(String listName)
    throws BadConnectionException, NoSuchDataException, TooManyValuesException
  {
    throw new NoSuchDataException("not implemented");
  }

  //---------------------------------------------------------------------------

  /**
   * Gets DeviceList object from a database query string.
   * Devices are grouped by class. Ordering of groups and devices is
   * implementation-dependent.
   *
   * @param query  a selection statement depending on the implementation
   * @return array of device list names
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception BadQueryException      -
   */
  public DeviceList getDeviceListByQuery(String query)
    throws BadConnectionException, NoSuchDataException, BadQueryException
  {
    throw new NoSuchDataException("not implemented");
  }


  //===========================================================================
  // DeviceProperty
  //===========================================================================

  /**
   * Gets DeviceProperty object for given property name and class name.
   *
   * @param  className the name of the class
   * @param  propertyName the property name for the class
   * @return DeviceProperty object
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException -
   */
  public DeviceProperty getDeviceProperty(String propertyName,
					  String className)
    throws BadConnectionException, NoSuchDataException, TooManyValuesException
  {
    throw new NoSuchDataException("not implemented");
  }

  //---------------------------------------------------------------------------

  /**
   * Instantiates common DeviceProperty object for property name and device class names.
   * This determines what can be done collectively with a set of devices
   * belonging to different classes, for instance a group from a DeviceList.
   * <P>
   * To be compatible, the properties must:
   * <UL>
   * <LI> have the same datatype
   * <LI> be all scalar or all array
   * <LI> have the same service name
   * </UL>
   * Throw NoSuchDataException if the classes are incompatible for the property.
   * <P>
   * The values of the composite property for the classes are :
   * <UL>
   * <LI> dataSize is the largest size
   * <LI> constantProp is false if one of them is false
   * <LI> writeProp is false if one of them is false
   * <LI> needCycleType is true if one of them is true
   * <LI> description is the first non null description
   * </UL>
   * <P>
   * Only the compatible characteristics defined in all the classes are listed.
   * To be compatible, characteristics which have a property name as value
   * must have the same name for all the classes. Characteristics which have
   * different values for some of the classes are given the value string
   * "MULTIVALUED"
   *
   * @param propertyName  the name of the property
   * @param classNames the classes for which the property should be defined
   * @return DeviceProperty object
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   */
  public DeviceProperty getCommonProperty(String propertyName,
					  String[] classNames)
    throws BadConnectionException, NoSuchDataException
  {
    throw new NoSuchDataException("not implemented");
  }


  //===========================================================================
  // DeviceMessage
  //===========================================================================

  /**
   * Gets DeviceMessage object corresponding to level, number and service.
   * If the serviceName is irrelevant, set it to "ALL".
   *
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException - when duplicate messageKey found.
   */
  public DeviceMessage getDeviceMessage(int messageLevel,
					int messageNumber, String serviceName)
    throws BadConnectionException, NoSuchDataException, TooManyValuesException
  {
    throw new NoSuchDataException("not implemented");
  }


  //===========================================================================
  // BitPatternDefinition
  //===========================================================================

  /**
   * Gets BitPatternDefinition object for the given pattern name.
   *
   * @param  patternName the pattern name for the class
   * @return BitPatternDefinition object
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException -
   */
  public BitPatternDefinition getBitPatternDefinition(String patternName)
    throws BadConnectionException, NoSuchDataException, TooManyValuesException
  {
    throw new NoSuchDataException("not implemented");
  }


  //===========================================================================
  // Symbols
  //===========================================================================

  /**
   * Gets the symbol value when given the symbol category and symbol name.
   *
   * @param symbolCategory the category of the symbol
   * @param symbolName     the name of the symbol
   * @return the corresponding symbol value
   * @exception BadConnectionException -
   * @exception NoSuchDataException    -
   * @exception TooManyValuesException -
   */
  public int getSymbolValue(String symbolCategory, String symbolName)
    throws BadConnectionException, NoSuchDataException, TooManyValuesException
  {
    throw new NoSuchDataException("not implemented");
  }

  //---------------------------------------------------------------------------

  private Data parse(String query) {
    String cname = null;
    String dname = null;
    Data data = new Data();

    StringTokenizer st = new StringTokenizer(query,"\t\n\r=",true);
    while (st.hasMoreElements()) {
      String tok = st.nextToken();
      if (tok.equals("class")) {
	if (st.nextToken().equals("=")) cname = st.nextToken();
      } else if (tok.equals("device")) {
	if (st.nextToken().equals("=")) dname = st.nextToken();
      }
    }
    if (cname!=null || dname!=null) {
      if (cname != null) data.insert("class",cname);
      if (dname != null) data.insert("device",dname);
    } else {
      data.insert("device",query); // if no tags found, assume just device spec
    }
    return data;
  }
}
