//-----------------------------------------------------------------------------
//
//  Description:
//      Event class for cdev. 
//
//  Author: Chip Watson
//
//-----------------------------------------------------------------------------
// Revision History:
//
// 11-Aug-98  CW Prototype version
//  3-Sep-98  FDM added getOpcode
//  8-Sep-98  CW moved coco and errmsg into a DeviceError reference
//  5-Nov-98  CW added state field and isComplete() method
//  9-Nov-98  FDM modified toString
//  2-Dec-98  M.C modified paramString()
// 12-Fev-99  FDM added facilitating functions for end users
//  3-Jun-99  CW allow inheritance s.t. ClipTransaction does not use Device
//  3-Feb-00  CW bug fix in paramString (in case device name null)
//
//-----------------------------------------------------------------------------

package cdev;

import cdev.data.*;

import java.util.*;

public class DeviceEvent extends EventObject {

  protected Device device;
  protected int opcode;
  protected String arg; // cdev message or property name
  protected Data data;
  protected DeviceError error;
  protected int state; // filled in by ControlSystem

  // -----------------------------------------------------------------------------------
  /**
   * Objects of this class are constructed by CDEV and underlying services to convey
   * data and error information to client programs. (Never constructed directly).
   * @param device    Device generating the event
   * @param opcode    Operation generating the event (enumeration in DeviceTransaction)
   * @param arg       Either the property name or message to send (depends on opcode)
   * @param data      Outbound data
   * @param error     Error associated with this event, or null
   */
  // -----------------------------------------------------------------------------------
  public DeviceEvent(Device device, int opcode, String arg, Data data, DeviceError error)
  {
    super(DeviceEvent.class); // not used, but must specify a "source"
    this.device = device;
    this.opcode = opcode;
    this.arg = arg;
    this.data = data;
    this.error = error;
  }

  public void setSource(Object o) {source = o;}

  public Device getDevice() {return device;}

  public String getMessage() {
    if (opcode == DeviceTransaction.SEND)
      return arg;
    return null;
  }

  public String getProperty() {
    if (opcode == DeviceTransaction.SEND)
      return null;
    return arg;
  }

  public String getDeviceName() {   // must not be final
    if (device != null)
      return device.getName();
    else
      return null;
  }

  public String [] getDeviceNames() { // must not be final
    if (device != null) {
      if (device instanceof DeviceArray) {
	String [] temp = null;
	try {
	  temp = ((DeviceArray)device).getNames();
	} catch (Exception e) {}
	return temp;
      } else {
	String [] names = new String[1];
	names[0] = device.getName();
	return names;
      }
    } else
      return null;
  }

  public final DeviceError getError() {return error;}

  public final Data getData() {return data;}

  public final int getOpcode() { return opcode;}

  public final boolean isComplete() { return (state == DeviceTransaction.COMPLETE);}


  // Convenience functions (request from PS users , Jan 1999

  /** Returns the cycle type (CERN-PS: PLS line name). */
  public String getCycleType() {
      if (data != null) {
	  DataEntry de = data.getDataEntry("cycleType");
	  if (de != null)
	      return de.stringValue();
      }
      return null;
  }

  /** Returns the cycle stamp (CERN-PS: super-cycle number (16 bits) + basic period number (16 bits). */
  public int getCycleStamp() {
      if (data != null) {
	  DataEntry de = data.getDataEntry("cycleStamp");
	  if (de != null)
	      return de.intValue();
      }
      return 0;
  }

  /** Returns the value DataEntry */
  public DataEntry getValue() {
      return data.getDataEntry("value");
  }

  protected  String paramString() {
    StringBuffer str = new StringBuffer(getDeviceName());
    switch (opcode) {
    case DeviceTransaction.GET: str.append(",get"); break;
    case DeviceTransaction.SET: str.append(",set"); break;
    case DeviceTransaction.MONITORON: str.append(",monitorOn"); break;
    case DeviceTransaction.MONITOROFF: str.append(",monitorOff"); break;
    case DeviceTransaction.SEND: str.append(",send"); break;
    case DeviceTransaction.GETREFERENCE: str.append(",getReference"); break;
    case DeviceTransaction.SETREFERENCE: str.append(",setReference"); break;
    default: str.append(",unknown opcode "+opcode); break;
    }
    str.append(",").append(arg);
    return str.toString();
  }

  public String toString() {
    StringBuffer sb = new StringBuffer(getClass().getName());
    sb.append("[").append(paramString()).append("]");
    return sb.toString();
    // No data or error information in the string.
  }

}

  
